# CardDrawProbabilities.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Dialog to display deck analysis software
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>,
# GPL - see COPYING for details
"""Calculate probabilities for drawing the current selection."""

import gtk
from copy import copy
from sutekh.core.SutekhObjects import PhysicalCardSet, IAbstractCard
from sutekh.SutekhUtility import is_crypt_card
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint_error, \
        do_complaint_warning
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow


def _choose(iChoices, iTotal):
    """Returns number of unordered combinations of iChoices objects from
       iTotal (iTotal)(iTotal-1)...(iTotal-iChoices+1)/iChoices!"""
    if iChoices > 0:
        iDenom = iChoices
    else:
        return 1  # 0!/0! = 1, since 0! = 1
    iNumerator = iTotal
    for iNum in range(1, iChoices):
        iNumerator *= (iTotal - iNum)
        iDenom *= iNum
    return iNumerator / iDenom


def _gen_choice_list(dSelectedCounts):
    """Recursively generate all possible choices"""
    aList = []
    aSelectOrder = sorted(dSelectedCounts.items(), key=lambda x: (x[1], x[0]),
            reverse=True)
    sThisItem = aSelectOrder[0][0]
    if len(dSelectedCounts) > 1:
        dNew = copy(dSelectedCounts)
        del dNew[sThisItem]
        aSubLists = _gen_choice_list(dNew)
        for iChoice in range(dSelectedCounts[sThisItem] + 1):
            for aChoices in aSubLists:
                aThisList = [iChoice]
                aThisList.extend(aChoices)
                aList.append(aThisList)
    else:
        aList = [[x] for x in range(dSelectedCounts[sThisItem] + 1)]
    aList.sort(key=lambda x: [sum(x)] + x)
    return aList


def _multi_hyper_prob(aFound, iDraws, aObjects, iTotal):
    """Multivariate hypergeometric probability:

       Given a list of draw numbers: aFound = [iFound1, iFound2 ... iFoundN]
       form aObjects = [iObjects1, iObjects2 ... iObjectsN]
       return the probably of seeing exactly aFound from iDraws
       """
    # Complain about impossible cases
    if len(aFound) != len(aObjects):
        raise RuntimeError('Invalid input: aFound : %s, aObjects: %s' % (
            ','.join([str(x) for x in aFound]), ','.join([str(x) for x in
                aObjects])))
    if sum(aFound) < 0 or min(aFound) < 0 or sum(aObjects) < 0 or \
            iTotal <= 0 or iDraws <= 0 or min(aObjects) < 0 or \
            sum(aObjects) > iTotal or iDraws > iTotal:
        raise RuntimeError('Invalid values for multivariate hypergeomtric'
                ' probability calculation: aFound: %s iDraws: %d'
                ' aObjects: %s iTotal: %d' % (','.join([str(x) for x in
                    aFound]), iDraws, ','.join([str(x) for x in aObjects]),
                    iTotal))
    # Eliminate trivial cases
    if sum(aFound) > iDraws:
        return 0.0
    for iFound, iObjects in zip(aFound, aObjects):
        if iFound > iObjects:
            return 0.0
    # Hypergeomteric probability: P(X = iFound) = choose iFound from iObjects *
    #           choose (iDraws - iFound) from (iTotal - iObjects) /
    #           choose iDraws from iTotal
    # Multivariate: P(X_i = iFound[i]) = choose(iFound1, iObjects1) *
    #           choose(iFound2, iObject2) * ... / choose(iDraws, iTotal)
    fDemon = float(_choose(iDraws, iTotal))
    # Ensure we handle last entry
    iRemObjects = iTotal - sum(aObjects)
    iRemFound = iDraws - sum(aFound)
    fNumerator = 1.0
    for iFound, iObjects in zip(aFound, aObjects):
        fNumerator *= _choose(iFound, iObjects)
    if iRemFound > 0 and iRemObjects > 0:
        fNumerator *= _choose(iRemFound, iRemObjects)
    return fNumerator / fDemon


def _hyper_prob_at_least(aFound, iDraws, aObjects, iTotal, iCurCol=0):
    """Returns the probablity of drawing at least aFound from aObjects objects
       of interest from iTotal objects in iDraw draws."""
    # This is the sum of hyper_prob(iCur, Draws, iObjects, iTotal) where
    # iCur in [iFound, min(iDraws, iObjects)]
    if len(aFound) != len(aObjects):
        raise RuntimeError('Invalid input: aFound : %s, aObjects: %s' % (
            ','.join([str(x) for x in aFound]), ','.join([str(x) for x in
                aObjects])))
    fProb = 0
    aThisFound = copy(aFound)
    for iCur in range(aFound[iCurCol], min(iDraws, aObjects[iCurCol]) + 1):
        aThisFound[iCurCol] = iCur
        if iCurCol < len(aFound) - 1:
            fProb += _hyper_prob_at_least(aThisFound, iDraws,
                    aObjects, iTotal, iCurCol + 1)
        else:
            fProb += _multi_hyper_prob(aThisFound, iDraws, aObjects, iTotal)
    return fProb


class CardDrawSimPlugin(SutekhPlugin):
    """Displays the probabilities for drawing cards from the current
       selection."""
    # pylint: disable-msg=R0902
    # we use a lot of attributes to pass the data around
    dTableVersions = {PhysicalCardSet: (4, 5, 6)}
    aModelsSupported = (PhysicalCardSet,)

    def get_menu_item(self):
        """Register on the 'Analyze' menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oCardDraw = gtk.MenuItem("Card Draw probabilities")
        oCardDraw.connect("activate", self.activate)
        return ('Analyze', oCardDraw)

    # pylint: disable-msg=W0201
    # W0201 - we define lots of things here, rather than __init__, since this
    # is the plugin's entry point, and they need to reflect the current state
    def activate(self, _oWidget):
        """Create the actual dialog, and populate it."""
        if not self.check_cs_size('Card Draw probabilities', 500):
            return

        self.iTotal = 0
        self.dSelectedCounts = {}
        self.iSelectedCount = 0

        # Get currently selected cards
        aSelectedCards, bCrypt, bLibrary = self._get_selected_cards()

        if len(aSelectedCards) == 0:
            do_complaint_error("No cards selected")
            return

        # Check that selection doesn't mix crypt and libraries
        if bLibrary and bCrypt:
            do_complaint_error("Can't operate on selections including both"
                    " Crypt and Library cards")
            return

        self._setup_cardlists(aSelectedCards, bCrypt)

        # Query user if we have zero counts
        if 0 in self.dSelectedCounts.values():
            iRes = do_complaint_warning("Selected cards include cards with"
                    " a count of 0.\nDo you want to continue?")
            if iRes == gtk.RESPONSE_CANCEL:
                return

        oMainTitle = gtk.Label()
        if bCrypt:
            oMainTitle.set_markup('<b>Crypt:</b> Drawing from <b>%d</b> '
                    'cards' % self.iTotal)
            self.iOpeningDraw = 4
            self.iNumSteps = min(2, self.iTotal - self.iOpeningDraw)
        else:
            oMainTitle.set_markup('<b>Library:</b> Drawing from <b>%d</b> '
                    'cards' % self.iTotal)
            self.iOpeningDraw = 7
            # Look 15 cards into the deck by default, seems good start
            self.iNumSteps = min(8, self.iTotal - self.iOpeningDraw)
        self.iMax = min(15, self.iTotal - self.iOpeningDraw)
        self.iDrawStep = 1  # Increments to use in table
        self.aAllChoices = _gen_choice_list(self.dSelectedCounts)
        self.iCardsToDraw = min(3, self.iSelectedCount)

        if self.iTotal <= self.iOpeningDraw:
            if bLibrary:
                do_complaint_error("Library must be larger than the opening"
                        " hand")
            else:
                do_complaint_error("Crypt must be larger than"
                        " the opening draw")
            return

        self.oResultsTable = gtk.Table()
        oDialog = self._make_dialog(oMainTitle, bCrypt)
        self._fill_table(self, bCrypt)
        oDialog.set_size_request(800, 600)
        oDialog.show_all()

        oDialog.run()

    def _make_dialog(self, oMainTitle, bCrypt):
        """Create the dialog box."""
        oDialog = SutekhDialog("Card Draw probablities", self.parent,
            gtk.DIALOG_DESTROY_WITH_PARENT,
            (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        oDialog.connect("response", lambda oDialog, resp: oDialog.destroy())

        oWidgetBox = gtk.HBox(False, 2)
        oNumDraws = gtk.combo_box_new_text()
        iIndex = 0
        for iNum in range(1, self.iMax + 1):
            oNumDraws.append_text(str(iNum))
            if iNum < self.iNumSteps:
                iIndex += 1
        oNumDraws.set_active(iIndex)
        oNumDraws.connect('changed', self._cols_changed)

        oStepSize = gtk.combo_box_new_text()
        for iNum in range(1, 11):
            oStepSize.append_text(str(iNum))
        oStepSize.set_active(0)
        oStepSize.connect('changed', self._steps_changed)

        iIndex = 0
        oCardToDrawCount = gtk.combo_box_new_text()
        for iNum in range(1, self.iSelectedCount + 1):
            oCardToDrawCount.append_text(str(iNum))
            if iNum < self.iCardsToDraw:
                iIndex += 1
        oCardToDrawCount.set_active(iIndex)
        oCardToDrawCount.connect('changed', self._rows_changed)

        oRecalcButton = gtk.Button('recalculate table')

        oRecalcButton.connect('clicked', self._fill_table, bCrypt)

        oWidgetBox.pack_start(gtk.Label('columns in table : '))
        oWidgetBox.pack_start(oNumDraws)
        oWidgetBox.pack_start(gtk.Label(' Step between columns : '))
        oWidgetBox.pack_start(oStepSize)
        oWidgetBox.pack_start(gtk.Label(' cards of interest : '))
        oWidgetBox.pack_start(oCardToDrawCount)
        oWidgetBox.pack_start(gtk.Label(' : '))
        oWidgetBox.pack_start(oRecalcButton)

        oDescLabel = gtk.Label("Columns are number of draws (X)\n"
                "rows are the number of cards (Y)\n"
                "For the first row (0 cards), the entry is the possiblity of "
                "not drawing any of the selected cards by the X'th draw\n"
                "For the other rows, the entries are the probability of "
                "seeing at least the listed cards from the selection\n"
                "by the X\'th draw with the number in brackets being the "
                "probablity of seeing eactly the listed cards\n"
                "The first column represents the opening hand or crypt draw")

        oTitleLabel = gtk.Label('Draw results :')
        oResultsBox = gtk.VBox(False, 2)
        oResultsBox.pack_start(oTitleLabel, False, False)

        # pylint: disable-msg=E1101
        # pylint doesn't detect gtk methods correctly

        oResultsBox.pack_start(AutoScrolledWindow(self.oResultsTable,
            True))

        oDialog.vbox.pack_start(oMainTitle, False, False)
        oDialog.vbox.pack_start(gtk.Label('Card Probablities'), False, False)
        oDialog.vbox.pack_start(oDescLabel, False, False)
        oDialog.vbox.pack_start(oWidgetBox, False, False)
        oDialog.vbox.pack_start(oResultsBox)
        return oDialog

    def _setup_cardlists(self, aSelectedCards, bCrypt):
        """Extract the needed card info from the model"""
        aAllAbsCards = [IAbstractCard(oCard) for oCard in
                self.get_all_cards()]
        iCryptSize = 0
        iLibrarySize = 0
        self.dSelectedCounts = {}
        self.iSelectedCount = 0
        # Initialise dict 1st, as cards with a count of 0 in the selection
        # are possible.
        for oCard in aSelectedCards:
            self.dSelectedCounts.setdefault(oCard, 0)
        for oCard in aAllAbsCards:
            if is_crypt_card(oCard):
                iCryptSize += 1
            else:
                iLibrarySize += 1
            if oCard in self.dSelectedCounts:
                self.dSelectedCounts[oCard] += 1
                self.iSelectedCount += 1
            # The assumption is that the user is interested in all copies of
            # the selected cards (as it seems the most useful), so we treat
            # the selection of any instance of the card in the list as
            # selecting all of them
        if bCrypt:
            self.iTotal = iCryptSize
        else:
            self.iTotal = iLibrarySize

    def _get_selected_cards(self):
        """Extract selected cards from the selection."""
        aSelectedCards = []
        bCrypt = False
        bLibrary = False
        _oModel, aSelection = self.view.get_selection().get_selected_rows()
        for oPath in aSelection:
            # pylint: disable-msg=E1101
            # pylint doesn't pick up adaptor's methods correctly
            oCard = IAbstractCard(self.model.get_card_name_from_path(oPath))
            if is_crypt_card(oCard):
                bCrypt = True
            else:
                bLibrary = True
            aSelectedCards.append(oCard)

        return aSelectedCards, bCrypt, bLibrary

    def _steps_changed(self, oComboBox):
        """Handle changes to the oStepSize combo box."""
        self.iDrawStep = int(oComboBox.get_active_text())

    def _rows_changed(self, oComboBox):
        """Handle changes to the oCardToDrawCount combo box."""
        self.iCardsToDraw = int(oComboBox.get_active_text())

    def _cols_changed(self, oComboBox):
        """Handle changes to the oNumDraws combo box."""
        self.iNumSteps = int(oComboBox.get_active_text())

    def _fill_table(self, _oWidget, bCrypt):
        """Fill Results Box with the draw results.

           oWidget is a dummy placeholder so this method can be called by the
           'connect' signal.
           """
        # This is messy, but does the job
        iNumCardRows = len([sum(x) for x in self.aAllChoices if
                sum(x) <= self.iCardsToDraw])
        iNumRows = 2 * iNumCardRows + 5
        if len(self.aAllChoices[0]) > 1:
            iNumCols = 2 * self.iNumSteps + 5
            iOffset = 2
        else:
            iNumCols = 2 * self.iNumSteps + 3
            iOffset = 0
        self._setup_table(iNumRows, iNumCols)
        if iOffset > 0:
            self.oResultsTable.attach(gtk.VSeparator(), 3, 4, 2, iNumRows,
                    xpadding=0, ypadding=0, xoptions=gtk.FILL,
                    yoptions=gtk.FILL)
            oLabel = gtk.Label('Count')
            self.oResultsTable.attach(oLabel, 2, 3, 2, 3)
            self._setup_count_rows()
        if bCrypt:
            oLabel = gtk.Label('Opening Draw')
        else:
            oLabel = gtk.Label('Opening Hand')
        self.oResultsTable.attach(oLabel, 4 + iOffset, 5 + iOffset,
                2, 3)
        self.oResultsTable.attach(gtk.HSeparator(), iOffset, iNumCols, 1,
                2, xpadding=0, ypadding=0, xoptions=gtk.FILL,
                yoptions=gtk.FILL)
        self.oResultsTable.attach(gtk.VSeparator(), 1, 2, 1, iNumRows,
                xpadding=0, ypadding=0, xoptions=gtk.FILL, yoptions=gtk.FILL)
        self.oResultsTable.attach(gtk.VSeparator(), 3 + iOffset, 4 + iOffset,
                2, iNumRows, xpadding=0, ypadding=0, xoptions=gtk.FILL,
                yoptions=gtk.FILL)
        for iCol in range(1, self.iNumSteps):
            oLabel = gtk.Label('+ %d' % (iCol * self.iDrawStep))
            iTableCol = 2 * iCol + 3 + iOffset
            self.oResultsTable.attach(gtk.VSeparator(), iTableCol,
                    iTableCol + 1, 2, iNumRows, xpadding=0, ypadding=0,
                    xoptions=gtk.FILL, yoptions=gtk.FILL)
            self.oResultsTable.attach(oLabel, iTableCol + 1, iTableCol + 2,
                    2, 3)

        aSelectOrder = sorted(self.dSelectedCounts.items(),
                key=lambda x: (x[1], x[0]), reverse=True)
        for iRow in range(iNumCardRows):
            oLabel = gtk.Label(self._gen_row_label(iRow, aSelectOrder))
            iTableRow = 2 * iRow + 3
            self.oResultsTable.attach(gtk.HSeparator(), 1 + iOffset,
                    iNumCols, iTableRow, iTableRow + 1, xpadding=0, ypadding=0,
                    xoptions=gtk.FILL, yoptions=gtk.FILL)
            self.oResultsTable.attach(oLabel, 2 + iOffset, 3 + iOffset,
                    iTableRow + 1, iTableRow + 2)
        # Fill in zero row
        aCardCounts = [x[1] for x in aSelectOrder]
        self._fill_row(0, iOffset, True, aCardCounts)
        # Fill in other rows
        for iRow in range(1, iNumCardRows):
            self._fill_row(iRow, iOffset, False, aCardCounts)
        self.oResultsTable.show_all()

    def _setup_table(self, iNumRows, iNumCols):
        """Clear the table, and setup constant entries."""
        for oChild in self.oResultsTable.get_children():
            self.oResultsTable.remove(oChild)
        self.oResultsTable.resize(iNumRows, iNumCols)
        self.oResultsTable.set_col_spacings(0)
        self.oResultsTable.set_row_spacings(0)
        oTopLabel = gtk.Label('Cards drawn')
        oLeftLabel = gtk.Label('Number of cards seen')
        oLeftLabel.set_angle(270)
        self.oResultsTable.attach(oTopLabel, 0, iNumCols, 0, 1)
        self.oResultsTable.attach(oLeftLabel, 0, 1, 1, iNumRows)

    def _setup_count_rows(self):
        """Fill in the cumulative count rows."""
        iBottomRow = 3
        for iCardRow in range(self.iCardsToDraw + 1):
            self.oResultsTable.attach(gtk.HSeparator(), 1, 3, iBottomRow,
                    iBottomRow + 1, xpadding=0, ypadding=0, xoptions=gtk.FILL,
                    yoptions=gtk.FILL)
            oLabel = gtk.Label('%d' % iCardRow)
            iNumRows = len([x for x in self.aAllChoices if
                sum(x) == iCardRow])
            iTopRow = iBottomRow + 2 * iNumRows
            self.oResultsTable.attach(oLabel, 2, 3, iBottomRow + 1, iTopRow)
            iBottomRow = iTopRow

    def _fill_row(self, iRow, iOffset, bZero, aCardCounts):
        """Fill a single row of the results table"""
        iTableRow = 2 * iRow + 4
        aThisDraw = self._gen_draw(iRow)
        for iCol in range(self.iNumSteps):
            iNumDraws = iCol * self.iDrawStep + self.iOpeningDraw
            if iNumDraws < self.iTotal:
                fProbExact = _multi_hyper_prob(aThisDraw, iNumDraws,
                        aCardCounts, self.iTotal) * 100
                if not bZero:
                    fProbAccum = _hyper_prob_at_least(aThisDraw, iNumDraws,
                            aCardCounts, self.iTotal) * 100
                    oResLabel = gtk.Label('%3.2f (%3.2f)' % (fProbAccum,
                        fProbExact))
                else:
                    oResLabel = gtk.Label('%3.2f' % fProbExact)
            else:
                if bZero:
                    oResLabel = gtk.Label('all cards drawn')
                else:
                    oResLabel = gtk.Label('')
            iTableCol = 2 * iCol + 4 + iOffset
            self.oResultsTable.attach(oResLabel, iTableCol, iTableCol + 1,
                    iTableRow, iTableRow + 1)

    def _gen_draw(self, iRow):
        """Construct the card draw for the row"""
        return self.aAllChoices[iRow]

    def _gen_row_label(self, iRow, aSelectOrder):
        """Construct the label."""
        if iRow == 0:
            return 'None of the cards'
        aThisRow = self._gen_draw(iRow)
        sLabel = ''
        for oItem, iCount in zip(aSelectOrder, aThisRow):
            sLabel += u'%d \u00D7 %s\n' % (iCount, oItem[0].name)
        return sLabel

plugin = CardDrawSimPlugin



# ARDBXMLInvParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# ARDB XML Inventory File Parser
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Parser for ARDB XML inventory format."""

# pylint: disable-msg=E0611, F0401
# pylint doesn't like the handling of the differences between 2.4 and 2.5
try:
    from xml.etree.ElementTree import XMLParser
except ImportError:
    from elementtree.ElementTree import XMLTreeBuilder as XMLParser
# For compatability with ElementTree 1.3
try:
    from xml.etree.ElementTree import ParseError
except ImportError:
    from xml.parsers.expat import ExpatError as ParseError
# pylint: enable-msg=E0611, F0401
from sutekh.core.SutekhObjects import csv_to_canonical
from sutekh.core.ArdbInfo import unescape_ardb_expansion_name


class ARDBInvXMLState(object):
    """Simple State tracker used by the XMLParser"""
    # tag states of interest
    # We try and honour set info, although current ARDB seems a bit odd in
    # how it sets this
    ROOTTAG, NOTAG, INCARD, CARDNAME, CARDSET, ADVANCED = range(6)

    COUNT_KEY = 'have'

    ROOT = 'inventory'

    def __init__(self, oHolder):
        self._sData = ""
        self._sCardName = None
        self._sCardSet = None
        self._iCount = 0
        self._iState = self.ROOTTAG
        self._oHolder = oHolder
        self._sAdvanced = ''

    def reset(self):
        """Reset internal state"""
        self._sData = ""
        self._sCardName = None
        self._sCardSet = None
        self._iCount = 0
        self._iState = self.ROOTTAG
        self._sAdvanced = ''

    def start(self, sTag, dAttributes):
        """Start tag encountered"""
        if self._iState == self.ROOTTAG:
            if sTag == self.ROOT:
                self._iState = self.NOTAG
            else:
                raise IOError('Not a ARDB %s XML file type' % self.ROOT)
        elif self._iState == self.INCARD:
            if sTag == 'name':
                self._iState = self.CARDNAME
            elif sTag == 'set':
                self._iState = self.CARDSET
            elif sTag == 'adv':
                self._iState = self.ADVANCED
        elif self._iState == self.NOTAG:
            if sTag == 'vampire' or sTag == 'card':
                self._iState = self.INCARD
                self._iCount = int(dAttributes[self.COUNT_KEY])

    def end(self, sTag):
        """End tag encountered"""
        if self._iState == self.INCARD:
            if sTag == 'vampire' or sTag == 'card':
                if not self._sCardSet:
                    # convert empty string to None
                    self._sCardSet = None
                else:
                    self._sCardSet = unescape_ardb_expansion_name(
                            self._sCardSet)
                if self._sAdvanced == '(Advanced)':
                    self._sCardName = self._sCardName + ' (Advanced)'
                self._sCardName = csv_to_canonical(self._sCardName)
                self._oHolder.add(self._iCount, self._sCardName,
                        self._sCardSet)
                self._sCardName = None
                self._sCardSet = None
                self._iState = self.NOTAG
                self._sAdvanced = ''
        elif self._iState == self.CARDNAME and sTag == 'name':
            self._iState = self.INCARD
            self._sCardName = self._sData
            self._sData = ""
        elif self._iState == self.CARDSET and sTag == 'set':
            self._iState = self.INCARD
            self._sCardSet = self._sData
            self._sData = ""
        elif self._iState == self.ADVANCED and sTag == 'adv':
            self._iState = self.INCARD
            self._sAdvanced = self._sData
            self._sData = ""

    def data(self, sText):
        """Text data for current tag"""
        if self._iState not in [self.INCARD, self.NOTAG]:
            if self._sData:
                self._sData += sText
            else:
                self._sData = sText


class ARDBXMLInvParser(object):
    """Parser for the ARDB Inventory XML format."""
    _cState = ARDBInvXMLState

    def parse(self, fIn, oHolder):
        """Parse XML file into the card set holder"""
        oParser = XMLParser(target=self._cState(oHolder))
        try:
            for sLine in fIn:
                oParser.feed(sLine)
        except ParseError, oExp:
            raise IOError('Not an XML file: %s' % oExp)



# ConfigFile.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Config handling object
# Wrapper around configobj and validate with some hooks for Sutekh purposes
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2010 Simon Cross <hodgestar+sutekh@gmail.com>
# License: GPL - See COPYRIGHT file for details

"""Configuration handling for the Sutekh GUI."""

from configobj import ConfigObj, flatten_errors
from validate import Validator, is_option, is_list
import pkg_resources
import weakref

# Type definitions
CARDSET = 'Card Set'
FRAME = 'Frame'
WW_CARDLIST = 'cardlist'
CARDSET_LIST = 'cardset list'


def is_option_list(sValue, *aOptions):
    """Validator function for option_list configspec type."""
    return [is_option(sMem, *aOptions) for sMem in is_list(sValue)]


class ConfigFileListener(object):
    """Listener object for config changes - inspired by CardListModeListener"""

    def add_filter(self, sKey, sFilter):
        """New Filter added"""
        pass

    def remove_filter(self, sKey, sFilter):
        """Filter removed"""
        pass

    def replace_filter(self, sKey, sOldFilter, sNewFilter):
        """A Filter has been replaced"""
        pass

    def set_postfix_the_display(self, bPostfix):
        """Postfix display mode has been set."""
        pass

    def profile_option_changed(self, sType, sProfile, sKey):
        """One of the profile configuration items changed."""
        pass

    def profile_changed(self, sType, sId):
        """The profile associated with a cardset changed."""
        pass

    def remove_profile(self, sType, sProfile):
        """A profile has been removed"""
        pass


class ConfigFile(object):
    """Handle the setup and management of the config file.

       Ensure that the needed sections exist, and that sensible
       defaults values are assigned.

       Filters are saved to the config file, and interested objects
       can register as listeners on the config file to respond to
       changes to the filters.
       """
    # pylint: disable-msg=R0904, R0902
    # R0904 - We need to provide fine-grained access to all the data,
    # so lots of methods
    # R0902 - Lots of internal state, so lots of attributes

    dCustomConfigTypes = {
        'option_list': is_option_list,
    }

    def __init__(self, sFileName):
        self.__sFileName = sFileName
        self.__bWriteable = False
        self.__oConfigSpec = None
        self.__oConfig = None
        self.__dLocalFrameOptions = {}
        self.__oValidator = None
        # allow listeners to be automatically removed by garbage
        # collection
        self.__dListeners = weakref.WeakKeyDictionary()
        self.__dPluginSpecs = {}
        self.__dDeckSpecs = {}
        self.__dCardListSpecs = {}
        self.__dCardSetListSpecs = {}

    def __str__(self):
        """Debugging aid - include the filename"""
        return "<%s object at %s; config file: %r>" % (
            self.__class__.__name__, hex(id(self)), self.__sFileName)

    def add_listener(self, oListener):
        """Add a listener to respond to config file changes."""
        self.__dListeners[oListener] = None

    def remove_listener(self, oListener):
        """Remove a listener from the list."""
        del self.__dListeners[oListener]

    def listeners(self):
        """Return a list of listeners.

        Useful for iterating over the listeners since the internal
        list is stored in a weakref dictionary.
        """
        return self.__dListeners.keys()

    def add_plugin_specs(self, sName, dConfigSpecs):
        """Add a validator to the plugins_main configspec section."""
        self.__dPluginSpecs[sName] = dConfigSpecs

    def add_deck_specs(self, sName, dConfigSpecs):
        """Add validation options to the per_deck.defaults configspec."""
        self.__dDeckSpecs[sName] = dConfigSpecs

    def add_cardlist_specs(self, sName, dConfigSpecs):
        """Add validation options to the cardlist configspec."""
        self.__dCardListSpecs[sName] = dConfigSpecs

    def add_cardset_list_specs(self, sName, dConfigSpecs):
        """Add validation options to the cardset list configspec."""
        self.__dCardSetListSpecs[sName] = dConfigSpecs

    def validate(self):
        """Validate a configuration object."""
        # pylint: disable-msg=E1101
        # pkg_resources confuses pylint here
        fConfigSpec = pkg_resources.resource_stream(__name__, "configspec.ini")
        oConfigSpec = ConfigObj(fConfigSpec, raise_errors=True,
                file_error=True, list_values=False)

        for sPlugin, dGlobal in self.__dPluginSpecs.items():
            oConfigSpec['plugins_main'][sPlugin] = dGlobal

        for sPlugin, dPerDeck in self.__dDeckSpecs.items():
            for sKey, oValue in dPerDeck.items():
                oConfigSpec['per_deck']['defaults'][sKey] = oValue

        for sPlugin, dGlobal in self.__dCardListSpecs.items():
            for sKey, oValue in dGlobal.items():
                oConfigSpec['cardlist']['defaults'][sKey] = oValue

        for sPlugin, dGlobal in self.__dCardSetListSpecs.items():
            for sKey, oValue in dGlobal.items():
                oConfigSpec['cardset list']['defaults'][sKey] = oValue

        # Set __many__ sections to match defaults sections, so validation
        # does the value conversion for us

        for sKey, oValue in oConfigSpec['cardlist']['defaults'].items():
            oConfigSpec['cardlist']['profiles']['__many__'][sKey] = oValue
        for sKey, oValue in oConfigSpec['cardset list']['defaults'].items():
            oConfigSpec['cardset list']['profiles']['__many__'][sKey] = oValue
        for sKey, oValue in oConfigSpec['per_deck']['defaults'].items():
            oConfigSpec['per_deck']['profiles']['__many__'][sKey] = oValue

        self.__oConfigSpec = oConfigSpec
        self.__oConfig = ConfigObj(self.__sFileName, configspec=oConfigSpec,
                indent_type='    ')
        self.__oValidator = Validator(self.dCustomConfigTypes)

        oResults = self.__oConfig.validate(self.__oValidator,
                preserve_errors=True)
        return oResults

    def validation_errors(self, oValidationResults):
        """Return a list of string describing any validation errors.

           Returns an empty list if no validation errors occurred. Validation
           results must have been returned by a call to validate() on the same
           config object.
           """
        aErrors = []
        if oValidationResults == True:
            return aErrors

        for (aSections, sKey, _oIgnore) in flatten_errors(self.__oConfig,
            oValidationResults):
            if sKey is not None:
                aErrors.append('Key %r in section %r failed validation.' %
                        (sKey, aSections))
            else:
                aErrors.append('Section %r was missing.' % (aSections,))

        return aErrors

    def get_validator(self):
        """Return the validator used to check the configuration."""
        return self.__oValidator

    def sanitize(self):
        """Called after validation to clean up a valid config.

           Currently clean-up consists of adding some open panes if none
           are listed.
           """
        if not self.__oConfig['open_frames']:
            # No panes information, so we set 'sensible' defaults
            self.add_frame(1, 'physical_card', 'White Wolf Card List', False,
                    False, -1, None)
            self.add_frame(2, 'Card Text', 'Card Text', False, False, -1, None)
            self.add_frame(3, 'Card Set List', 'Card Set List', False, False,
                    -1, None)
            self.add_frame(4, 'physical_card_set', 'My Collection', False,
                    False, -1, None)

    def check_writeable(self):
        """Test that we can open the file for writing"""
        self.__bWriteable = True
        try:
            oFile = open(self.__sFileName, 'a')
            oFile.close()
        except IOError:
            self.__bWriteable = False
        return self.__bWriteable

    def write(self):
        """Write the config file to disk."""
        if self.__bWriteable:
            self.__oConfig.write()

    #
    # Open Frame Handling
    #

    def clear_open_frames(self):
        """Clear out old save panes (used before adding new ones)."""
        self.__oConfig['open_frames'].clear()

    def open_frames(self):
        """Get the all the panes saved in the config file, and their
           positions."""
        aRes = []
        for sKey, dPane in self.__oConfig['open_frames'].items():
            if not sKey.startswith('pane'):
                # invalid key
                continue
            try:
                iPaneNumber = int(sKey.split(' ')[1])
            except ValueError:
                # invalid key
                continue

            iPos = dPane['position']
            if dPane['orientation'] == 'C':
                # We aim for minimal impact on the layout of other panes
                iPos = 0

            aRes.append((
                iPaneNumber,
                dPane['type'],
                dPane['name'],
                dPane['paneid'],
                dPane['orientation'] == 'V',
                dPane['orientation'] == 'C',
                iPos,
            ))

        aRes.sort()  # Numbers denote ordering
        return aRes

    def update_pane_numbers(self, dPaneMap):
        """Update the profiles to reflect changes in pane-id.

           dPaneMap is a dictionary mapping old_id -> new_id"""
        # We use a copy, so we avoid problems where sNewId is equal to
        # a later sOldId
        dOldProfiles = self.__oConfig['per_deck']['frame_profiles'].copy()
        # Clear old state, ensuring we set any botched loads to the
        # default profile
        for sId in dOldProfiles:
            self.set_profile(FRAME, sId, None)
        for sOldId, sNewId in dPaneMap.iteritems():
            if sOldId in dOldProfiles:
                # use set_profile to ensure we call the listeners
                self.set_profile(FRAME, sNewId, dOldProfiles[sOldId])

    # pylint: disable-msg=R0913
    # We need all the info in the arguments here
    def add_frame(self, iFrameNumber, sType, sName, bVertical, bClosed, iPos,
            sPaneId):
        """Add a frame with the given position info to the config file"""
        oPanes = self.__oConfig['open_frames']
        sKey = 'pane %d' % iFrameNumber

        oPanes[sKey] = {}
        oNewPane = oPanes[sKey]

        oNewPane['type'] = sType
        oNewPane['name'] = sName
        oNewPane['paneid'] = sPaneId

        if bVertical:
            oNewPane['orientation'] = "V"
        elif bClosed:
            oNewPane['orientation'] = "C"
        else:
            oNewPane['orientation'] = "H"

        if iPos > 0 and self.get_save_precise_pos() and not bClosed:
            oNewPane['position'] = iPos
        elif bClosed:
            oNewPane['position'] = 0
        else:
            oNewPane['position'] = -1

    # pylint: enable-msg=R0913

    #
    # Plugin Config Section Handling
    #

    def get_plugin_key(self, sPlugin, sKey):
        """Get an option from the plugins section.

           Return None if no option is set
           """
        try:
            sResult = self.__oConfig['plugins_main'][sPlugin][sKey]
        except KeyError:
            sResult = None
        return sResult

    def set_plugin_key(self, sPlugin, sKey, sValue, bCreateSection=False):
        """Set a value in the plugin section"""
        if bCreateSection and sPlugin not in self.__oConfig['plugins_main']:
            self.__oConfig['plugins_main'][sPlugin] = {}
        self.__oConfig['plugins_main'][sPlugin][sKey] = sValue

    #
    # Filters section
    #

    def get_filters(self):
        """Get all the filters in the config file.

           Filters are return as a list of (sQuery, dVars) tuples.
           """
        return [(oF['query'], oF['vars']) for oF in
            self.__oConfig['filters'].values()]

    def get_filter_keys(self):
        """Return all the keys for all the filters in the config file."""
        return [(sKey, dFilter['query']) for sKey, dFilter in
            self.__oConfig['filters'].items()]

    def get_filter(self, sKey):
        """Return the filter associated with the given key.

           Return None if the filter is not known.
           """
        sKey = sKey.lower()
        if sKey in self.__oConfig['filters']:
            return self.__oConfig['filters'][sKey]['query']
        else:
            return None

    def get_filter_values(self, sKey):
        """Return the filter values associated with the given key.

        Return None if the filter is not knonw.
        """
        sKey = sKey.lower()
        if sKey in self.__oConfig['filters']:
            return self.__oConfig['filters'][sKey]['vars'].dict()
        else:
            return None

    # pylint: disable-msg=W0102
    # W0102 - {} is the right thing here
    def add_filter(self, sKey, sQuery, dVars={}):
        """Add a filter to the config file."""
        sKey = sKey.lower()
        if sKey not in self.__oConfig['filters']:
            dFilter = {
                'query': sQuery,
                'vars': dVars,
            }
            self.__oConfig['filters'][sKey] = dFilter
            for oListener in self.listeners():
                oListener.add_filter(sKey, sQuery)

    # pylint: enable-msg=W0102

    def remove_filter(self, sKey, sFilter):
        """Remove a filter from the file"""
        # Make sure Filer and key match
        sKey = sKey.lower()
        if sKey in self.__oConfig['filters'] and \
                sFilter == self.__oConfig['filters'][sKey]['query']:
            del self.__oConfig['filters'][sKey]
            for oListener in self.listeners():
                oListener.remove_filter(sKey, sFilter)

    def replace_filter(self, sKey, sOldFilter, sNewFilter):
        """Replace a filter in the file with new filter"""
        sKey = sKey.lower()
        if sKey in self.__oConfig['filters'] and \
                sOldFilter == self.__oConfig['filters'][sKey]['query']:
            self.__oConfig['filters'][sKey]['query'] = sNewFilter
            for oListener in self.listeners():
                oListener.replace_filter(sKey, sOldFilter, sNewFilter)

    #
    # Per-Deck Options
    #

    def get_deck_option(self, sFrame, sCardset, sKey, bUseLocal=True):
        """Retrieve the value of a per-deck option.

           Either sFrame or sCardset may be None, in
           which case the frame or cardset option look-up
           is skipped.
           """
        try:
            if bUseLocal:
                return self.__dLocalFrameOptions[sFrame][sKey]
        except KeyError:
            pass

        dPerDeck = self.__oConfig['per_deck']

        try:
            sProfile = dPerDeck['frame_profiles'][sFrame]
            return dPerDeck['profiles'][sProfile][sKey]
        except KeyError:
            pass

        try:
            sProfile = dPerDeck['cardset_profiles'][sCardset]
            return dPerDeck['profiles'][sProfile][sKey]
        except KeyError:
            pass

        return dPerDeck['defaults'][sKey]

    def get_deck_profile_option(self, sProfile, sKey):
        """Get the value of a per-deck option for a profile."""
        try:
            if sProfile is None:
                return self.__oConfig['per_deck']['defaults'][sKey]
            else:
                return self.__oConfig['per_deck']['profiles'][sProfile][sKey]
        except KeyError:
            return None

    def get_local_frame_option(self, sFrame, sKey):
        """Get the value of a per-deck option for a local frame."""
        try:
            return self.__dLocalFrameOptions[sFrame][sKey]
        except KeyError:
            return None

    def set_profile_option(self, sType, sProfile, sKey, sValue):
        """Set the value of a option for a profile for sType.

           If sValue is None, remove the key. New profiles are
           created as needed.
           """
        if sType == CARDSET:
            dConfig = self.__oConfig['per_deck']
        else:
            dConfig = self.__oConfig[sType]
        if sProfile is None:
            dProfile = dConfig['defaults']
        elif sProfile in dConfig['profiles']:
            dProfile = dConfig['profiles'][sProfile]
        else:
            # configobj replaces the dict with a config object, so
            # trigger the creation of the config object, then set
            # dProfile to it
            dConfig['profiles'][sProfile] = {}
            dProfile = dConfig['profiles'][sProfile]

        bChanged = False
        if sValue is None:
            if sKey in dProfile:
                bChanged = True
                del dProfile[sKey]
        else:
            if sKey not in dProfile or dProfile[sKey] != sValue:
                bChanged = True
                dProfile[sKey] = sValue

        if bChanged:
            for oListener in self.listeners():
                oListener.profile_option_changed(sType, sProfile, sKey)

    def set_local_frame_option(self, sFrame, sKey, sValue):
        """Set the value of an option in the local frame option dictionary.

        If sValue is None, remove the key.
        """
        if sFrame in self.__dLocalFrameOptions:
            dOptions = self.__dLocalFrameOptions[sFrame]
        else:
            dOptions = {}
            self.__dLocalFrameOptions[sFrame] = dOptions

        bChanged = False
        if sValue is None:
            if sKey in dOptions:
                bChanged = True
                del dOptions[sKey]
        else:
            if sKey not in dOptions or dOptions[sKey] != sValue:
                bChanged = True
                dOptions[sKey] = sValue

        if bChanged:
            for oListener in self.listeners():
                oListener.profile_changed(FRAME, sFrame)

    def clear_frame_profile(self, sId):
        """Clear any pane profiles set for this frame"""
        dProfiles = self.__oConfig['per_deck']['frame_profiles']
        if sId in dProfiles:
            # We should only be called while the pane is being closed,
            # so we don't notify the listeners to avoid causing a
            # pointless reload.
            del dProfiles[sId]

    def clear_cardset_profile(self, sId):
        """Clear any profiles set for this cardset"""
        # We don't need to worry about the pane profiles, as the
        # the frames will be explicitly closed, which will take care of those
        # for us.
        dProfiles = self.__oConfig['per_deck']['cardset_profiles']
        if sId in dProfiles:
            # We should only be called while the pane is being closed,
            # so we don't notify the listeners to avoid causing a
            # pointless reload.
            del dProfiles[sId]

    def fix_profile_mapping(self, dOldMap, dNewMap):
        """Update the card set profiles to a new id -> name mapping"""
        # We need to reverse of dNewMapping. Since both name and id are
        # unique, this is safe
        dNewRev = dict(zip(dNewMap.itervalues(), dNewMap.iterkeys()))
        dOldProfiles = self.__oConfig['per_deck']['cardset_profiles'].copy()
        dProfiles = self.__oConfig['per_deck']['cardset_profiles']
        dProfiles.clear()
        for sId, sProfile in dOldProfiles.iteritems():
            iId = int(sId[2:])  # cs<cardset id>
            if iId not in dOldMap:
                # ini file out of synce with database, so skip this entry
                # (lots of ways this can happen - sutekh-cli, editing, etc.)
                continue
            sName = dOldMap[iId]
            if sName in dNewRev:
                sNewId = 'cs%d' % dNewRev[sName]
                dProfiles[sNewId] = sProfile

    def set_profile(self, sType, sId, sProfile):
        """Set the profile associated of the given type."""
        if sType == CARDSET or sType == FRAME:
            if sType == CARDSET:
                dProfiles = self.__oConfig['per_deck']['cardset_profiles']
            else:
                dProfiles = self.__oConfig['per_deck']['frame_profiles']
            if dProfiles.get(sId) == sProfile:
                return
            if sProfile is None:
                del dProfiles[sId]
            else:
                dProfiles[sId] = sProfile
        elif (sType == WW_CARDLIST and sId == WW_CARDLIST) or \
                (sType == CARDSET_LIST and sId == CARDSET_LIST):
            sCurProfile = self.__oConfig[sType].get('current profile')
            if sCurProfile == sProfile:
                return
            if sProfile is None:
                del self.__oConfig[sType]['current profile']
            else:
                self.__oConfig[sType]['current profile'] = sProfile
        else:
            # Unrecognised type / cardset combo
            return
        for oListener in self.listeners():
            oListener.profile_changed(sType, sId)

    def get_profile(self, sType, sId):
        """Return the current profile of the cardset/cardlist/cardset list."""
        if sType == CARDSET:
            return self.__oConfig['per_deck']['cardset_profiles'].get(sId)
        elif sType == FRAME:
            return self.__oConfig['per_deck']['frame_profiles'].get(sId)
        elif (sType == WW_CARDLIST and sId == WW_CARDLIST) or \
                (sType == CARDSET_LIST and sId == CARDSET_LIST):
            return self.__oConfig[sType].get('current profile')
        return None

    def clear_profiles(self, sType, sProfile):
        """Find all cardsets/frames using this profile, and change them to
          the default"""
        if sType == CARDSET or sType == FRAME:
            if sType == CARDSET:
                dProfiles = self.__oConfig['per_deck']['cardset_profiles']
            else:
                dProfiles = self.__oConfig['per_deck']['frame_profiles']
            # Loop over a copy, as we may change the dict
            for sId, sCurProfile in dProfiles.items():
                if sCurProfile == sProfile:
                    self.set_profile(sType, sId, None)
        else:
            sCurProfile = self.__oConfig[sType].get('current profile')
            if sCurProfile == sProfile:
                self.set_profile(sType, sType, None)

    def remove_profile(self, sType, sProfile):
        """Remove a profile from the file"""
        dData = {}
        if sType == FRAME or sType == CARDSET:
            dData = self.__oConfig['per_deck']['profiles']
        elif sType == CARDSET_LIST:
            dData = self.__oConfig['cardset list']['profiles']
        elif sType == WW_CARDLIST:
            dData = self.__oConfig['cardlist']['profiles']
        if sProfile in dData:
            if sType == FRAME or sType == CARDSET:
                # Need to clear both lists in this case
                self.clear_profiles(FRAME, sProfile)
                self.clear_profiles(CARDSET, sProfile)
            else:
                self.clear_profiles(sType, sProfile)
            del dData[sProfile]
            for oListener in self.listeners():
                oListener.remove_profile(sType, sProfile)

    def frame_profiles(self):
        """Return a dictionary of frame id -> profile mappings."""
        return dict(self.__oConfig['per_deck']['frame_profiles'])

    def get_profile_users(self, sType, sProfile):
        """Returns a list of all card sets or panes that use the
           given profile"""
        if sType == CARDSET_LIST:
            sCurProfile = self.__oConfig[sType].get('current profile')
            if sCurProfile == sProfile:
                return ['Card Set List']
        elif sType == WW_CARDLIST:
            sCurProfile = self.__oConfig[sType].get('current profile')
            if sCurProfile == sProfile:
                return ['White Wolf Card List']
        elif sType == CARDSET:
            aUsers = []
            for dProfiles in (
                    self.__oConfig['per_deck']['cardset_profiles'],
                    self.__oConfig['per_deck']['frame_profiles'],
                    ):
                for sId, sCurProfile in dProfiles.iteritems():
                    if sProfile == sCurProfile:
                        aUsers.append(sId)
            return aUsers
        return None

    def profiles(self, sType):
        """Return a list of profile keys."""
        if sType == FRAME or sType == CARDSET:
            return list(self.__oConfig['per_deck']['profiles'].keys())
        elif sType == CARDSET_LIST:
            return list(self.__oConfig['cardset list']['profiles'].keys())
        elif sType == WW_CARDLIST:
            return list(self.__oConfig['cardlist']['profiles'].keys())
        # Unkown type
        return None

    def profile_options(self, sType):
        """Return a list of per-deck option names."""
        if sType == FRAME or sType == CARDSET:
            return self.__oConfig['per_deck']['defaults'].keys()
        return self.__oConfig[sType]['defaults'].keys()

    def get_option_spec(self, sType, sKey):
        """Return the config spec for a given option."""
        if sType == FRAME or sType == CARDSET:
            return self.__oConfigSpec['per_deck']['defaults'][sKey]
        return self.__oConfigSpec[sType]['defaults'][sKey]

    def get_profile_option(self, sType, sProfile, sKey):
        """Get the value of a per-deck option for a profile."""
        if sType == FRAME or sType == CARDSET:
            return self.get_deck_profile_option(sProfile, sKey)
        try:
            if sProfile is None or sProfile.lower() == "default":
                return self.__oConfig[sType]['defaults'][sKey]
            else:
                # We fall back to the defaults if the key isn't set yet
                try:
                    return self.__oConfig[sType]['profiles'][sProfile][sKey]
                except KeyError:
                    return self.__oConfig[sType]['defaults'][sKey]
        except KeyError:
            return None

    #
    # Application Level Config Settings
    #

    def get_save_on_exit(self):
        """Query the 'save on exit' option."""
        return self.__oConfig['main']['save on exit']

    def set_save_on_exit(self, bSaveOnExit):
        """Set the 'save on exit' option."""
        self.__oConfig['main']['save on exit'] = bSaveOnExit

    def get_save_precise_pos(self):
        """Query the 'save pane sizes' option."""
        return self.__oConfig['main']['save pane sizes']

    def set_save_precise_pos(self, bSavePos):
        """Set the 'save pane sizes' option."""
        self.__oConfig['main']['save pane sizes'] = bSavePos

    def get_save_window_size(self):
        """Query the 'save window size' option."""
        return self.__oConfig['main']['save window size']

    def set_save_window_size(self, bSavePos):
        """Set the 'save window size' option."""
        self.__oConfig['main']['save window size'] = bSavePos

    def set_database_uri(self, sDatabaseURI):
        """Set the configured database URI"""
        self.__oConfig['main']['database url'] = sDatabaseURI

    def get_database_uri(self):
        """Get database URI from the config file"""
        return self.__oConfig['main']['database url']

    def get_icon_path(self):
        """Get the icon path from the config file"""
        return self.__oConfig['main']['icon path']

    def set_icon_path(self, sPath):
        """Set the configured icon path"""
        self.__oConfig['main']['icon path'] = sPath

    def get_window_size(self):
        """Get the saved window size from the config file."""
        iWidth, iHeight = self.__oConfig['main']['window size']
        return iWidth, iHeight

    def set_window_size(self, tSize):
        """Save the current window size."""
        self.__oConfig['main']['window size'] = tSize

    def get_postfix_the_display(self):
        """Get the 'postfix name display' option."""
        return self.__oConfig['main']['postfix name display']

    def set_postfix_the_display(self, bPostfix):
        """Set the 'postfix name display' option."""
        self.__oConfig['main']['postfix name display'] = bPostfix
        for oListener in self.listeners():
            oListener.set_postfix_the_display(bPostfix)



# FilterEditor.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Editor component for generic filter ASTs.
# Copyright 2008 Simon Cross <hodgestar@gmail.com>
# Copyright 2010 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Widget for editing filters."""

from sutekh.gui.SutekhDialog import SutekhDialog
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.FilterModelPanes import FilterModelPanes, add_accel_to_button
from sutekh.core.FilterParser import get_filters_for_type
import gtk
import pango


class FilterEditor(gtk.Alignment):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """GTK component for editing Sutekh filter ASTs.

       Provides a graphical representation of the filter as nested boxes,
       which the user can extend or remove from.
       """
    def __init__(self, oAST, sFilterType, oParser, oFilterDialog):
        # Child widget absorbs all free space
        super(FilterEditor, self).__init__(xscale=1.0, yscale=1.0)
        self.__oParser = oParser
        self.__oFilterDialog = oFilterDialog  # Dialog we're a child of

        self.__oPanes = FilterModelPanes(sFilterType, oFilterDialog)

        self.__sFilterType = sFilterType

        oNameLabel = gtk.Label("Filter name:")
        self.__oNameEntry = gtk.Entry()
        self.__oNameEntry.set_width_chars(30)

        oHelpButton = gtk.Button("Help")
        oHelpButton.connect("clicked", self.__show_help_dialog)
        add_accel_to_button(oHelpButton, 'F1', oFilterDialog.accel_group)

        oHBox = gtk.HBox(spacing=5)
        oHBox.pack_start(oNameLabel, expand=False)
        oHBox.pack_start(self.__oNameEntry, expand=False)
        oHBox.pack_end(oHelpButton, expand=False)

        self.__oVBox = gtk.VBox(spacing=5)
        self.__oVBox.pack_end(oHBox, expand=False)
        self.__oVBox.pack_end(gtk.HSeparator(), expand=False)
        self.__oVBox.pack_start(self.__oPanes, expand=True)

        self.add(self.__oVBox)

        self.replace_ast(oAST)

    def get_filter(self):
        """Get the actual filter for the current AST."""
        oAST = self.get_current_ast()
        if oAST is None:
            return None
        else:
            return oAST.get_filter()

    def get_current_ast(self):
        """Get the current AST represented by the editor."""
        return self.__oPanes.get_ast_with_values()

    def get_current_text(self):
        """Get the current text of the filter for saving in the config
           file.

           We include any current set values for the filter when saving."""
        return self.__oPanes.get_text()

    def replace_ast(self, oAST):
        """Replace the current AST with a new one and update the GUI,
           preserving variable values if possible.

           Also used to setup the filter initially.
           """
        self.__oPanes.replace_ast(oAST)

    def set_name(self, sName):
        """Set the filter name."""
        self.__oNameEntry.set_text(sName)

    def get_name(self):
        """Get the filter name."""
        return self.__oNameEntry.get_text().strip()

    def connect_name_changed(self, fCallback):
        """Connect a callback to the name entry change signal."""
        self.__oNameEntry.connect('changed', fCallback)

    def __show_help_dialog(self, _oHelpButton):
        """Show a dialog window with the helptext from the filters."""
        oDlg = SutekhDialog("Help on Filters", self.__oFilterDialog,
                gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))

        oDlg.set_default_size(600, 500)
        # pylint: disable-msg=E1101
        # vbox confuse pylint

        # Ensure we sort this in the same order as the toolbar
        oHelpView = AutoScrolledWindow(FilterHelpTextView(
            sorted(get_filters_for_type(self.__sFilterType),
                key=lambda x: x.description)))
        oDlg.vbox.pack_start(oHelpView)
        oDlg.show_all()

        try:
            oDlg.run()
        finally:
            oDlg.destroy()


class FilterHelpBuffer(gtk.TextBuffer):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """TextBuffer object used to display the help about the filters."""
    def __init__(self):
        super(FilterHelpBuffer, self).__init__(None)

        # See http://www.pygtk.org/pygtk2reference/class-gtktexttag.html
        # for some possible properties

        self.create_tag("header", weight=pango.WEIGHT_BOLD,
                scale=pango.SCALE_LARGE, underline=pango.UNDERLINE_SINGLE)
        self.create_tag("description", weight=pango.WEIGHT_BOLD)
        self.create_tag("keyword", style=pango.STYLE_ITALIC)
        self.create_tag("helptext", left_margin=10)
        self._oIter = self.get_iter_at_offset(0)

    # pylint: disable-msg=W0142
    # ** magic OK here
    def tag_text(self, *aArgs, **kwargs):
        """Insert text into the buffer with tags."""
        self.insert_with_tags_by_name(self._oIter, *aArgs, **kwargs)

    # pylint: enable-msg=W0142

    def add_help_text(self, sDescription, sHelpText):
        """Add the given help text to the text buffer."""
        self.tag_text(sDescription, "description")
        self.tag_text(" :\n")
        self.tag_text(sHelpText, "helptext")
        self.tag_text("\n")

    def reset_iter(self):
        """reset the iterator to the start of the buffer."""
        self._oIter = self.get_iter_at_offset(0)


class FilterHelpTextView(gtk.TextView):
    """TextView widget for displaying the help text."""
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    def __init__(self, aFilterTypes):
        super(FilterHelpTextView, self).__init__()
        oBuf = FilterHelpBuffer()

        self.set_buffer(oBuf)
        self.set_editable(False)
        self.set_cursor_visible(False)
        self.set_wrap_mode(gtk.WRAP_WORD)
        self.set_border_width(5)
        self.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))

        oBuf.reset_iter()

        oBuf.tag_text('Usage', 'header')
        oBuf.tag_text("\n")

        oBuf.tag_text(
            "Filter elements and values for the filters can be dragged "
            "from the right hand pane to the filter to add them to "
            "the filter, and dragged from the filter to the right hand "
            "pane to remove them from the filter. Disabled filter elements "
            "or filter elements with no values set are ignored.\n\n",
            'helptext')

        oBuf.tag_text('Keyboard Navigation', 'header')
        oBuf.tag_text("\n")

        oBuf.tag_text("You can switch between the different filter lists "
                "using <Alt>-Left and <Alt>-Right. <Ctrl>-Enter will "
                "paste the currently selected list of values into the filter. "
                "<Del> will delete a value or filter part from the filter."
                "<Ctrl>-Space will disable a filter part, while <Alt>-Space "
                "will negate it.\n\n", "helptext")

        oBuf.tag_text('Available Filter Elements', 'header')
        oBuf.tag_text("\n")

        oBuf.tag_text("The available filtering options are listed below. "
            "The first line of each item shows the description "
            "you'll see in the filter editor in bold. "
            "The rest of the description describes the arguments "
            "the filter takes and the results it produces.\n\n", "helptext")

        # Add test for filter group
        oBuf.add_help_text('Filter Group', "Filter element with holds "
                "other filter elements. The result of is the combination "
                "of all the filter elements as specified by the Filter "
                "Group type (All of, Any of and so on).")

        for oFilt in aFilterTypes:
            oBuf.add_help_text(oFilt.description,
                    oFilt.helptext)

        oBuf.tag_text("\n"
            "Note: Using filters which require information about "
            "cards in a card set will temporarily turn off displaying "
            "cards with zero counts (for modes where this is possible) "
            "in the card list since there are no "
            "associated cards on which to base the query. "
            "Filters affected by this are: Physical Expansion, "
            "Card Sets and In Card Sets in Use.")



# ScrolledList.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Generic Scrolled List, used in the Filter Dialog and elsewhere"""

import gtk
import gobject
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.CustomDragIconView import CustomDragIconView


class ScrolledListStore(gtk.ListStore):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Simple list store for ScrolledList widget"""
    def __init__(self):
        super(ScrolledListStore, self).__init__(gobject.TYPE_STRING)

    def fill_list(self, aVals):
        """Fill the list"""
        self.clear()
        for sEntry in aVals:
            oIter = self.append(None)
            self.set(oIter, 0, sEntry)


class ScrolledListView(CustomDragIconView):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Simple tree view for the ScrolledList widget"""
    def __init__(self, sTitle, oModel=None, bSpecialSelect=False):
        if not oModel:
            oModel = ScrolledListStore()
        super(ScrolledListView, self).__init__(oModel)
        oCell1 = gtk.CellRendererText()
        oColumn1 = gtk.TreeViewColumn(sTitle, oCell1, markup=0)
        self.append_column(oColumn1)
        self._oSelection.set_mode(gtk.SELECTION_MULTIPLE)

        if bSpecialSelect:
            self._oSelection.connect('changed', self.row_selected)

    # pylint: disable-msg=W0212
    # allow access via these properties
    store = property(fget=lambda self: self._oModel, doc="List of values")
    # disable-msg=W0212

    def get_selected_data(self):
        """Get the list of selected values"""
        aSelectedList = []
        oModel, aSelectedRows = self._oSelection.get_selected_rows()
        for oPath in aSelectedRows:
            oIter = oModel.get_iter(oPath)
            sName = oModel.get_value(oIter, 0)
            aSelectedList.append(sName)
        return aSelectedList

    def set_selected_rows(self, aRowsToSelect):
        """Set the selection to the correct list"""
        aRowsToSelect = set(aRowsToSelect)
        oIter = self._oModel.get_iter_first()
        self._oSelection.unselect_all()
        while oIter is not None:
            sName = self._oModel.get_value(oIter, 0)
            if sName in aRowsToSelect:
                self._oSelection.select_iter(oIter)
            oIter = self._oModel.iter_next(oIter)

    def set_selected_entry(self, sEntry):
        """Set the selection to the correct entry"""
        oIter = self._oModel.get_iter_first()
        while oIter is not None:
            sName = self._oModel.get_value(oIter, 0)
            if sName == sEntry:
                oPath = self._oModel.get_path(oIter)
                self.set_cursor(oPath)
                return
            oIter = self._oModel.iter_next(oIter)


class ScrolledList(gtk.Frame):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Frame containing an auto scrolled list"""
    def __init__(self, sTitle, oModel=None, bSpecialSelect=None):
        super(ScrolledList, self).__init__(None)
        self._oTreeView = ScrolledListView(sTitle, oModel, bSpecialSelect)
        oMyScroll = AutoScrolledWindow(self._oTreeView)
        self.add(oMyScroll)
        self.set_shadow_type(gtk.SHADOW_NONE)
        self.show_all()

    # pylint: disable-msg=W0212
    # allow access via these properties
    view = property(fget=lambda self: self._oTreeView,
            doc="Associated View Object")
    # disable-msg=W0212

    def set_select_single(self):
        """set selection to single mode"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_SINGLE)

    def set_select_multiple(self):
        """set selection to multiple mode"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_MULTIPLE)

    def set_select_none(self):
        """set selection mode to none"""
        self._oTreeView.get_selection().set_mode(gtk.SELECTION_NONE)

    def get_selection(self):
        """Return a list of the selected elements of the list"""
        return self._oTreeView.get_selected_data()

    def set_selected_rows(self, aRowsToSelect):
        """Set the selected rows to aRowsToSelect."""
        self._oTreeView.set_selected_rows(aRowsToSelect)

    def set_selected_entry(self, sEntry):
        """Select the current entry"""
        self._oTreeView.set_selected_entry(sEntry)

    def fill_list(self, aVals):
        """Fill the list store with the given values"""
        self._oTreeView.store.fill_list(aVals)

    def get_selection_object(self):
        """Get a reference to the TreeView selection"""
        return self._oTreeView.get_selection()



# GuiCardSetFunctions.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Useful utilities for managing card sets that need to access the gui"""

import gtk
import datetime
import logging
from sutekh.core.CardSetHolder import CardSetHolder
from sutekh.gui.SutekhDialog import do_complaint_warning, do_complaint, \
        do_complaint_error, do_exception_complaint
from sutekh.core.SutekhObjects import PhysicalCardSet, IPhysicalCardSet
from sutekh.core.CardLookup import LookupFailed
from sutekh.gui.CreateCardSetDialog import CreateCardSetDialog
from sutekh.gui.RenameDialog import RenameDialog, PROMPT, RENAME, REPLACE
from sutekh.core.CardSetUtilities import delete_physical_card_set, \
        find_children, has_children, detect_loop, get_loop_names, break_loop


def reparent_card_set(oCardSet, oNewParent):
    """Helper function to ensure that reparenting a card set doesn't
       cause loops"""
    if oNewParent:
        # Can only be a problem if the new parent is a card set
        oOldParent = oCardSet.parent
        oCardSet.parent = oNewParent
        oCardSet.syncUpdate()
        if detect_loop(oCardSet):
            oCardSet.parent = oOldParent
            oCardSet.syncUpdate()
            do_complaint('Changing parent of %s to %s introduces a'
                    ' loop. Leaving the parent unchanged.' % (
                        oCardSet.name, oNewParent.name),
                    gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE)
        else:
            return True
    else:
        oCardSet.parent = oNewParent
        oCardSet.syncUpdate()
        return True
    return False


def reparent_all_children(sCardSetName, aChildren):
    """Handle reparenting a list of children gracefully"""
    if aChildren:
        oCardSet = IPhysicalCardSet(sCardSetName)
        for oChildCS in aChildren:
            reparent_card_set(oChildCS, oCardSet)


def check_ok_to_delete(oCardSet):
    """Check if the user is OK with deleting the card set."""
    bChildren = has_children(oCardSet)
    bCards = len(oCardSet.cards) > 0
    iResponse = gtk.RESPONSE_OK
    if bCards and bChildren:
        iResponse = do_complaint_warning("Card Set %s Not Empty and"
                " Has Children. Really Delete?" % oCardSet.name)
    elif bCards:
        iResponse = do_complaint_warning("Card Set %s Not Empty."
                " Really Delete?" % oCardSet.name)
    elif bChildren:
        iResponse = do_complaint_warning("Card Set %s"
                " Has Children. Really Delete?" % oCardSet.name)
    return iResponse == gtk.RESPONSE_OK


def create_card_set(oMainWindow):
    """Create a new card set from the edit dialog"""
    oDialog = CreateCardSetDialog(oMainWindow)
    oDialog.run()
    sName = oDialog.get_name()
    if sName:
        if PhysicalCardSet.selectBy(name=sName).count() != 0:
            do_complaint_error("Card Set %s already exists." % sName)
            return None
        sAuthor = oDialog.get_author()
        sComment = oDialog.get_comment()
        oParent = oDialog.get_parent()
        bInUse = oDialog.get_in_use()
        _oCS = PhysicalCardSet(name=sName, author=sAuthor,
                comment=sComment, parent=oParent, inuse=bInUse)
    return sName


# import helpers
def get_import_name(oHolder, iClashMode=PROMPT):
    """Helper for importing a card set holder.

       Deals with prompting the user for a new name if required, and properly
       dealing with child card sets if the user decides to replace an
       existing card set."""

    def setup_for_replace():
        """Little helper function to handle getting children and deleting
           the card set consistently"""
        oCS = IPhysicalCardSet(oHolder.name)
        aChildren = find_children(oCS)
        # Delete existing card set
        delete_physical_card_set(oHolder.name)
        return aChildren

    bRename = False
    if oHolder.name:
        # Check if we need to prompt for rename
        if PhysicalCardSet.selectBy(name=oHolder.name).count() != 0:
            bRename = True
    else:
        # No name, need to prompt
        bRename = True
    aChildren = []
    if bRename:
        if iClashMode == PROMPT:
            oDlg = RenameDialog(oHolder.name)
            iResponse = oDlg.run()
            if iResponse == RENAME:
                oHolder.name = oDlg.sNewName
            elif iResponse == REPLACE:
                # Get child card sets
                aChildren = setup_for_replace()
            else:
                # User cancelled, so bail
                oHolder.name = None
            oDlg.destroy()
        elif iClashMode == REPLACE:
            aChildren = setup_for_replace()
        elif iClashMode == RENAME:
             # Create a unique name based on our current name
            sTime = datetime.datetime.today().strftime('%Y-%m-%d %H:%M')
            sBaseName = '%s (imported %s)' % (oHolder.name, sTime)
            sNewName = sBaseName
            iCount = 0
            while PhysicalCardSet.selectBy(name=sNewName).count() != 0:
                iCount += 1
                sNewName = '%s (%d)' % (sBaseName, iCount)
            oHolder.name = sNewName

    return oHolder, aChildren


def update_open_card_sets(oMainWindow, sSetName):
    """Update open copies of the card set sSetName to database changes
       (from imports, etc.)"""
    for oFrame in oMainWindow.find_cs_pane_by_set_name(sSetName):
        oFrame.update_to_new_db()
    oMainWindow.reload_pcs_list()


def update_card_set(oCardSet, oMainWindow):
    """Update the details of the card set when the user edits them."""
    sOldName = oCardSet.name
    oEditDialog = CreateCardSetDialog(oMainWindow, oCardSet=oCardSet)
    oEditDialog.run()
    sName = oEditDialog.get_name()
    if not sName:
        return  # bail
    # To ensure we don't send unnecessary changed signals, only update
    # if things have changed
    if sName != oCardSet.name:
        oCardSet.name = sName
    sAuthor = oEditDialog.get_author()
    if sAuthor != oCardSet.author:
        oCardSet.author = sAuthor
    sComment = oEditDialog.get_comment()
    if sComment != oCardSet.comment:
        oCardSet.comment = sComment
    sAnnotations = oEditDialog.get_annotations()
    if sAnnotations != oCardSet.annotations:
        oCardSet.annotations = sAnnotations
    bInUse = oEditDialog.get_in_use()
    if bInUse != oCardSet.inuse:
        oCardSet.inuse = bInUse
    oParent = oEditDialog.get_parent()
    if oParent != oCardSet.parent:
        reparent_card_set(oCardSet, oParent)
    oCardSet.syncUpdate()
    # Update frame menu
    for oFrame in oMainWindow.find_cs_pane_by_set_name(sOldName):
        oFrame.menu.update_card_set_menu(oCardSet)
        # update_card_set_menu does the needed magic for us
    oMainWindow.reload_pcs_list()


# Common to MainMenu import code and plugins
def import_cs(fIn, oParser, oMainWindow, sSetName=None):
    """Create a card set from the given file object."""
    oHolder = CardSetHolder()

    # pylint: disable-msg=W0703
    # we really do want all the exceptions
    try:
        oParser.parse(fIn, oHolder)
    except Exception, oExp:
        sMsg = "Reading the card set failed with the following error:\n" \
               "%s\n The file is probably not in the format the Parser" \
               " expects.\nAborting" % oExp
        do_exception_complaint(sMsg)
        # Fail out
        return

    if oHolder.num_entries < 1:
        # No cards seen, so abort
        do_complaint_error("No cards found in the card set.\n"
                "The file may not be in the format the Parser expects.\n"
                "Aborting")
        return

    # Display any warnings
    aWarnings = oHolder.get_warnings()
    if aWarnings:
        sMsg = "The following warnings were reported:\n%s" % \
                "\n".join(aWarnings)
        logging.warn(sMsg)
        iResponse = do_complaint_warning("%s\nContinue with the import?"
                % sMsg)
        if iResponse != gtk.RESPONSE_OK:
            return  # bail out
        oHolder.clear_warnings()

    # Handle naming issues if needed
    if sSetName and not oHolder.name:
        # Use provided name if we don't have one from the set
        oHolder.name = sSetName
    oHolder, aChildren = get_import_name(oHolder)
    if not oHolder.name:
        return  # User bailed
    # Create CS
    try:
        oHolder.create_pcs(oCardLookup=oMainWindow.cardLookup)
        reparent_all_children(oHolder.name, aChildren)
        aWarnings = oHolder.get_warnings()
        if aWarnings:
            sMsg = "Card Set Created.\n" \
                    "The following warnings were reported during the final" \
                    " import:\n%s" % "\n".join(aWarnings)
            logging.warn(sMsg)
            do_complaint(sMsg, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE)
    except RuntimeError, oExp:
        sMsg = "Creating the card set failed with the following error:\n" \
               "%s\nAborting" % oExp
        do_exception_complaint(sMsg)
        return
    except LookupFailed, oExp:
        return

    if oMainWindow.find_cs_pane_by_set_name(oHolder.name):
        # Already open, so update to changes
        update_open_card_sets(oMainWindow, oHolder.name)
    else:
        # Not already open, so open a new copy
        oMainWindow.add_new_physical_card_set(oHolder.name)


def break_existing_loops():
    """Ensure there are no loops in the database"""
    for oCS in PhysicalCardSet.select():
        if detect_loop(oCS):
            sLoop = "->".join(get_loop_names(oCS))
            sBreakName = break_loop(oCS)
            do_complaint(
                    'Loop %s in the card sets relationships.\n'
                    'Breaking at %s' % (sLoop, sBreakName),
                    gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE)
            # We break the loop, and let the user fix things,
            # rather than try and be too clever



# AboutDialog.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Simple about dialog for Sutekh"""

from sutekh.gui import SutekhIcon
from sutekh.SutekhInfo import SutekhInfo
import gtk


# pylint: disable-msg=R0904
# R0904 - gtk Widget, so has many public methods
class SutekhAboutDialog(gtk.AboutDialog):
    """About dialog for Sutekh."""

    # pylint: disable-msg=W0142
    # ** magic OK here
    def __init__(self, *aArgs, **kwargs):
        super(SutekhAboutDialog, self).__init__(*aArgs, **kwargs)

        self.set_name(SutekhInfo.NAME)
        self.set_version(SutekhInfo.VERSION_STR)
        self.set_copyright(SutekhInfo.LICENSE)
        self.set_comments(SutekhInfo.DESCRIPTION)
        self.set_license(SutekhInfo.LICENSE_TEXT)
        self.set_wrap_license(False)  # don't automatically wrap license text
        self.set_website(SutekhInfo.SOURCEFORGE_URL)
        self.set_website_label("Website")
        self.set_authors([tAuth[0] for tAuth in SutekhInfo.AUTHORS])
        self.set_documenters([tAuth[0] for tAuth in SutekhInfo.DOCUMENTERS])
        self.set_artists([tAuth[0] for tAuth in SutekhInfo.ARTISTS])
        self.set_logo(SutekhIcon.SUTEKH_ICON)
        # self.set_translator_credits(translator_credits)



# ShowExportedCardSet.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2011 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Plugin for displaying the exported version of a card set in a gtk.TextView.
   Intended to make cutting and pasting easier."""

import gtk
import StringIO
from sutekh.core.SutekhObjects import PhysicalCardSet
from sutekh.core.CardSetHolder import CardSetWrapper
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow
from sutekh.gui.SutekhDialog import SutekhDialog
from sutekh.io.WriteJOL import WriteJOL
from sutekh.io.WriteLackeyCCG import WriteLackeyCCG
from sutekh.io.WriteELDBDeckFile import WriteELDBDeckFile
from sutekh.io.WriteArdbText import WriteArdbText
from sutekh.io.WriteVEKNForum import WriteVEKNForum
from sutekh.io.WriteCSV import WriteCSV


class ShowExported(SutekhPlugin):
    """Display the various exported versions of a card set.

       The card set is shown in a textview, and the user can toggle between
       the different formats. This is designed to make it trivial to
       cut-n-paste the card set into something else (such as a web-browser).
       """

    dTableVersions = {}
    aModelsSupported = (PhysicalCardSet,)

    _dExporters = {
            # radio button text : Writer
            'Export to JOL format': WriteJOL,
            'Export to Lackey CCG format': WriteLackeyCCG,
            'Export to ARDB Text': WriteArdbText,
            'BBcode output for the V:EKN Forums': WriteVEKNForum,
            'Export to ELDB ELD Deck File': WriteELDBDeckFile,
            'CSV Export (with headers)': WriteCSV,
            }

    def get_menu_item(self):
        """Register on the 'Analyze' menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oShowExported = gtk.MenuItem("Display card set in alternative format")
        oShowExported.connect("activate", self.activate)
        return ('Actions', oShowExported)

    def activate(self, _oWidget):
        """Handle response from menu"""
        oCardSet = self.get_card_set()
        if not oCardSet:
            return
        oDlg = SutekhDialog("Exported CardSet: %s" % self.view.sSetName,
                self.parent, gtk.DIALOG_DESTROY_WITH_PARENT)
        oDlg.set_default_size(700, 600)
        # Add scrolled window for text
        oTextBuffer = ExportBuffer()
        oTextView = gtk.TextView()
        oTextView.set_buffer(oTextBuffer)
        oTextView.set_editable(False)
        oTextView.set_wrap_mode(gtk.WRAP_NONE)  # preserve long lines
        oTextView.set_border_width(5)
        oTextView.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("white"))
        oDlg.vbox.pack_start(AutoScrolledWindow(oTextView))
        # Add the radio buttons
        oTable = gtk.Table(len(self._dExporters) / 2, 2)
        oIter = self._dExporters.iterkeys()
        iXPos, iYPos = 0, 0
        oFirstBut = None
        for sName in oIter:
            if not oFirstBut:
                oBut = gtk.RadioButton(None, sName)
                oFirstBut = oBut
                oFirstBut.set_active(True)
                self._set_text(sName, oCardSet, oTextBuffer)
            else:
                oBut = gtk.RadioButton(oFirstBut, sName)
            oBut.connect('toggled', self._button_toggled, sName, oCardSet,
                    oTextBuffer)
            oTable.attach(oBut, iXPos, iXPos + 1, iYPos, iYPos + 1)
            iXPos += 1
            if iXPos > 1:
                iXPos = 0
                iYPos += 1
        oDlg.vbox.pack_start(oTable, False)
        oDlg.add_button(gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE)
        oDlg.connect("response", lambda oW, oR: oDlg.destroy())
        oDlg.show_all()
        oDlg.run()

    def _button_toggled(self, oBut, sName, oCardSet, oTextBuffer):
        """Handle user changing the toggle button state"""
        if not oBut.get_active():
            # Only concerned with the button that's just become active
            return
        self._set_text(sName, oCardSet, oTextBuffer)

    def _set_text(self, sName, oCardSet, oTextBuffer):
        """Internals of setting the buffer to the correct text"""
        cWriter = self._dExporters[sName]
        oWriter = cWriter()
        fOut = StringIO.StringIO()
        oWriter.write(fOut, CardSetWrapper(oCardSet))
        oTextBuffer.set_text(fOut.getvalue())
        fOut.close()


class ExportBuffer(gtk.TextBuffer):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Buffer object for showing the exported card set text"""

    def __init__(self):
        super(ExportBuffer, self).__init__(None)
        self.create_tag("text", left_margin=0)

    def set_text(self, sCardSetText):
        """Set the buffer contents to the card set text"""
        oStart, oEnd = self.get_bounds()
        self.delete(oStart, oEnd)
        oIter = self.get_iter_at_offset(0)
        self.insert_with_tags_by_name(oIter, sCardSetText, "text")


plugin = ShowExported



# PhysicalCardView.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2005, 2006, 2007 Simon Cross <hodgestar@gmail.com>
# Copyright 2006, 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Provide a TreeView for the physical card collection"""

import gtk
from sutekh.gui.CardListView import CardListView
from sutekh.gui.CardListModel import CardListModel
from sutekh.gui.CellRendererIcons import CellRendererIcons


class PhysicalCardView(CardListView):
    # pylint: disable-msg=R0904, R0902, R0901
    # R0904 - gtk.Widget, so many public methods
    # R0902 - We need to track a fair amount of state, so many attributes
    # R0901 - many ancestors, due to our object hierachy on top of the quite
    # deep gtk one
    """The card list view for the physical card collection.

       Special cases Editable card list with those properties
       needed for the card collection - the drag prefix, the
       card_drop handling and handling of pasted data.
       """
    sDragPrefix = 'Phys:'

    def __init__(self, oController, oWindow, oConfig):
        oModel = CardListModel(oConfig)
        oModel.enable_sorting()
        super(PhysicalCardView, self).__init__(oController, oWindow,
                oModel, oConfig)

        # Setup columns for default view
        self.oNameCell = CellRendererIcons(5)

        oColumn = gtk.TreeViewColumn("Cards", self.oNameCell, text=0,
                textlist=5, icons=6)
        oColumn.set_expand(True)
        oColumn.set_resizable(True)
        oColumn.set_sort_column_id(0)
        self.append_column(oColumn)

        self.set_expander_column(oColumn)



# FullBackup.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details
"""Plugin to wrap zipfile backup and restore methods"""

from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import do_complaint_warning, \
        do_exception_complaint
from sutekh.gui.SutekhFileWidget import ZipFileDialog
from sutekh.gui.ProgressDialog import ProgressDialog, SutekhCountLogHandler
from sutekh.io.ZipFileWrapper import ZipFileWrapper
from sutekh.SutekhUtility import get_cs_id_name_table
import gtk
import os


class FullBackup(SutekhPlugin):
    """Provide access to ZipFileWrapper's backup and restore methods.

       Handle GUI aspects associated with restoring (ensuring everything
       reloads, etc.)
       """

    dTableVersions = {}
    aModelsSupported = ("MainWindow",)

    # Dialog and Menu Item Creation

    def get_menu_item(self):
        """Register on the Plugins menu"""
        if not self.check_versions() or not self.check_model_type():
            return None

        oBackup = gtk.MenuItem("Save a Full Backup")
        oBackup.connect("activate", self.activate_backup)
        oRestore = gtk.MenuItem("Restore a Full Backup")
        oRestore.connect("activate", self.activate_restore)

        return [('Backup', oBackup), ('Backup', oRestore)]

    # Menu responses

    def activate_backup(self, _oWidget):
        """Handle backup request"""
        oDlg = self.make_backup_dialog()
        oDlg.run()
        sFilename = oDlg.get_name()
        if sFilename:
            self.handle_backup_response(sFilename)

    def activate_restore(self, _oWidget):
        """Handle restore request"""
        oDlg = self.make_restore_dialog()
        oDlg.run()
        sFilename = oDlg.get_name()
        if sFilename:
            self.handle_restore_response(sFilename)

    # Backup

    def make_backup_dialog(self):
        """Create file dialog for backup"""
        sName = "Choose a file to save the full backup to ..."

        oDlg = ZipFileDialog(self.parent, sName, gtk.FILE_CHOOSER_ACTION_SAVE)
        oDlg.set_do_overwrite_confirmation(True)
        oDlg.show_all()

        return oDlg

    # pylint: disable-msg=R0201
    # This could be a function, but that won't add any clarity to this code
    def handle_backup_response(self, sFilename):
        """Handle response from backup dialog"""
        # pylint: disable-msg=W0703
        # we really do want all the exceptions
        try:
            oLogHandler = SutekhCountLogHandler()
            oProgressDialog = ProgressDialog()
            oProgressDialog.set_description("Saving backup")
            oLogHandler.set_dialog(oProgressDialog)
            oProgressDialog.show()
            oFile = ZipFileWrapper(sFilename)
            oFile.do_dump_all_to_zip(oLogHandler)
            oProgressDialog.destroy()
        except Exception, oException:
            oProgressDialog.destroy()
            sMsg = "Failed to write backup.\n\n%s" % oException
            do_exception_complaint(sMsg)

    # pylint: enable-msg=R0201

    # Restore

    def make_restore_dialog(self):
        """Create file chooser dialog for restore"""
        sName = "Restore a Full Backup ...."

        oDlg = ZipFileDialog(self.parent, sName, gtk.FILE_CHOOSER_ACTION_OPEN)

        oWarning = gtk.Label()
        oWarning.set_markup("<b>This will delete all existing Card Sets</b>")
        # pylint: disable-msg=E1101
        # plint doesn't pick up vbox methods correctly
        oDlg.vbox.pack_start(oWarning, expand=False)
        oDlg.vbox.reorder_child(oWarning, 0)
        oDlg.show_all()

        return oDlg

    def handle_restore_response(self, sFilename):
        """Handle response from the restore dialog"""
        bContinue = True

        if not os.path.exists(sFilename):
            bContinue = do_complaint_warning(
                    "Backup file %s does not seem to exist."
                    % sFilename) != gtk.RESPONSE_CANCEL

        if bContinue:
            dOldMap = get_cs_id_name_table()
            try:
                aEditable = self.parent.get_editable_panes()
                oLogHandler = SutekhCountLogHandler()
                oProgressDialog = ProgressDialog()
                oProgressDialog.set_description("Restoring backup")
                oLogHandler.set_dialog(oProgressDialog)
                oProgressDialog.show()
                oFile = ZipFileWrapper(sFilename)
                oFile.do_restore_from_zip(self.cardlookup, oLogHandler)
                # restore successful, refresh display
                aMessages = oFile.get_warnings()
                if aMessages:
                    sMsg = "The following warnings were reported:\n%s" % \
                            "\n".join(aMessages)
                    do_complaint_warning(sMsg)
                # Id's will not be preserved
                dNewMap = get_cs_id_name_table()
                self.parent.config_file.fix_profile_mapping(dOldMap, dNewMap)
                self.parent.update_to_new_db()
                oProgressDialog.destroy()
                self.parent.restore_editable_panes(aEditable)
            # pylint: disable-msg=W0703
            # we really do want all the exceptions
            except Exception, oException:
                oProgressDialog.destroy()
                sMsg = "Failed to restore backup.\n\n%s" % oException
                do_exception_complaint(sMsg)


plugin = FullBackup



# AnalyzeCardList.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Dialog to display deck analysis software
# Copyright 2006, 2007 Neil Muller <drnlmuller+sutekh@gmail.com>,
# Copyright 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details
# pylint: disable-msg=C0302
# C0302 - This covers a lot of cases, and splitting it into multiple
# files won't gain any clarity

"""Display interesting statistics and properties of the card set."""

import gtk
import logging
from sqlobject import SQLObjectNotFound
from sutekh.core.SutekhObjects import PhysicalCardSet, \
        IAbstractCard, IPhysicalCard, IExpansion, CRYPT_TYPES, IKeyword
from sutekh.core.Filters import CardTypeFilter, FilterNot
from sutekh.core.Abbreviations import Titles
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.SutekhDialog import SutekhDialog
from sutekh.gui.MultiSelectComboBox import MultiSelectComboBox

try:
    THIRD_ED = IExpansion('Third Edition')
    JYHAD = IExpansion('Jyhad')
except SQLObjectNotFound, oExcDetails:
    # log exception at same level as plugin errors (verbose)
    logging.warn("Expansion caching failed (%s).", oExcDetails, exc_info=1)
    # Turn exception into a ImportError, so plugin is just disabled
    raise ImportError('Unable to load the required expansions')
ODD_BACKS = (None, THIRD_ED, JYHAD)

UNSLEEVED = "<span foreground='green'>%s may be played unsleeved</span>\n"
SLEEVED = "<span foreground='orange'>%s should be sleeved</span>\n"
SPECIAL = set(['Not Tournament Legal Cards', 'Multirole', 'Mixed Card Backs'])


# utility functions
def _percentage(iNum, iTot, sDesc):
    """Utility function for calculating _percentages"""
    if iTot > 0:
        fPrec = iNum / float(iTot)
    else:
        fPrec = 0.0
    return '<i>(%5.3f %% of %s)</i>' % (fPrec * 100, sDesc)


def _get_abstract_cards(aCards):
    """Get the abstract cards given the list of names"""
    return [IAbstractCard(x) for x in aCards]


def _lookup_discipline(sKey, dDisciplines):
    """Return the object with the fullname sKey"""
    return [x for x in dDisciplines if sKey == x.fullname][0]


def _disc_sort_cmp(oTuple1, oTuple2):
    """Sort disciplines by reverse number, then reverse superior number,
       then alphabetically by name"""
    # Check numbers force, reversing cmp's results
    iCmp = cmp(oTuple1[1][1], oTuple2[1][1])
    if iCmp != 0:
        return -iCmp
    iCmp = cmp(oTuple1[1][2], oTuple2[1][2])
    if iCmp != 0:
        return -iCmp
    return cmp(oTuple1[0].fullname, oTuple2[0].fullname)


def _format_card_line(sString, sTrailer, iNum, iLibSize):
    """Format card lines for notebook"""
    sPer = _percentage(iNum, iLibSize, "Library")
    return "Number of %(type)s %(trail)s = %(num)d %(per)s\n" % {
            'type': sString,
            'trail': sTrailer,
            'num': iNum,
            'per': sPer, }


def _get_card_costs(aCards):
    """Calculate the cost of the list of Abstract Cards

       Return lists of costs, for pool, blood and convictions
       Each list contains: Number with variable cost, Maximum Cost, Total Cost,
       Number of cards with a cost
       """
    dCosts = {}
    for sType in ('blood', 'pool', 'conviction'):
        dCosts.setdefault(sType, [0, 0, 0, 0])
    for oAbsCard in aCards:
        if oAbsCard.cost is not None:
            dCosts[oAbsCard.costtype][3] += 1
            if oAbsCard.cost == -1:
                dCosts[oAbsCard.costtype][0] += 1
            else:
                iMaxCost = dCosts[oAbsCard.costtype][1]
                dCosts[oAbsCard.costtype][1] = max(iMaxCost, oAbsCard.cost)
                dCosts[oAbsCard.costtype][2] += oAbsCard.cost
    return dCosts['blood'], dCosts['pool'], dCosts['conviction']


def _get_card_disciplines(aCards):
    """Extract the set of disciplines and virtues from the cards"""
    dDisciplines = {}
    dVirtues = {}
    iNoneCount = 0
    for oAbsCard in aCards:
        if len(oAbsCard.discipline) > 0:
            aThisDisc = [oP.discipline.fullname for oP
                    in oAbsCard.discipline]
        else:
            aThisDisc = []
        if len(oAbsCard.virtue) > 0:
            aThisVirtue = [oV.fullname for oV in oAbsCard.virtue]
        else:
            aThisVirtue = []
        for sDisc in aThisDisc:
            dDisciplines.setdefault(sDisc, 0)
            dDisciplines[sDisc] += 1
        for sVirtue in aThisVirtue:
            dVirtues.setdefault(sVirtue, 0)
            dVirtues[sVirtue] += 1
        if len(oAbsCard.discipline) == 0 and len(oAbsCard.virtue) == 0:
            iNoneCount += 1
    return dDisciplines, dVirtues, iNoneCount


def _get_card_clan_multi(aCards):
    """Extract the clan requirements and the multi discipline cards
       form the list of Abstract Cards"""
    dClan = {}
    iClanRequirement = 0
    dMulti = {}
    for oAbsCard in aCards:
        if len(oAbsCard.clan) > 0:
            iClanRequirement += 1
            aClans = [x.name for x in oAbsCard.clan]
            for sClan in aClans:
                dClan.setdefault(sClan, 0)
                dClan[sClan] += 1
        aTypes = [x.name for x in oAbsCard.cardtype]
        if len(aTypes) > 1:
            sKey = "/".join(sorted(aTypes))
            dMulti.setdefault(sKey, 0)
            dMulti[sKey] += 1
    return iClanRequirement, dClan, dMulti


def _format_cost_numbers(sCardType, sCostString, aCost, iNum):
    """Format the display of the card cost information"""
    sVarPercent = _percentage(aCost[0], iNum, '%s cards' % sCardType)
    sNumPercent = _percentage(aCost[3], iNum, '%s cards' % sCardType)
    if iNum > 0:
        fAvg = aCost[2] / float(iNum)
    else:
        fAvg = 0.0
    sText = "Most Expensive %(name)s Card  (%(type)s) = %(max)d\n" \
            "Cards with variable cost = %(var)d %(per)s\n" \
            "Cards with %(type)s cost = %(numcost)d %(percost)s\n" \
            "Average %(name)s card %(type)s cost = %(avg)5.3f\n" % {
                    'name': sCardType,
                    'type': sCostString,
                    'var': aCost[0],
                    'per': sVarPercent,
                    'max': aCost[1],
                    'avg': fAvg,
                    'numcost': aCost[3],
                    'percost': sNumPercent,
                    }
    return sText


def _crypt_avg(iCostTot, iCryptSize):
    """Calculate the crypt averages"""
    if iCryptSize > 0:
        fAvg = float(iCostTot) / iCryptSize
    else:
        fAvg = 0.0
    if iCryptSize < 4:
        fAvgDraw = iCryptSize * fAvg
    else:
        fAvgDraw = 4 * fAvg

    return fAvg, fAvgDraw


def _format_disciplines(sDiscType, dDisc, iNum):
    """Format the display of disciplines and virtues"""
    sText = ''
    for sDisc, iNum in sorted(dDisc.items(), key=lambda x: x[1],
            reverse=True):
        sText += 'Number of cards requiring %s %s = %d\n' % (sDiscType,
                sDisc, iNum)
    return sText


def _format_clan(sCardType, iClanRequirement, dClan, iNum):
    """Format the clan requirements list for display"""
    sText = ''
    if iClanRequirement > 0:
        sPer = _percentage(iClanRequirement, iNum, '%s cards' % sCardType)
        sText += "Number of %(type)s with a Clan requirement = %(num)d " \
                "%(per)s\n" % {'type': sCardType, 'num': iClanRequirement,
                'per': sPer, }
        for sClan, iClanNum in sorted(dClan.items(), key=lambda x: x[1],
                reverse=True):
            sPer = _percentage(iClanNum, iNum, '%s cards' % sCardType)
            sText += 'Number of %(type)s requiring %(clan)s = %(num)d ' \
                    '%(per)s\n' % {'type': sCardType, 'clan': sClan,
                    'num': iClanNum, 'per': sPer, }
    return sText


def _format_multi(sCardType, dMulti, iNum):
    """Format the multi-role cards list for display"""
    sText = ''
    for sType, iMulti in sorted(dMulti.items(), key=lambda x: x[1],
            reverse=True):
        sPer = _percentage(iMulti, iNum, '%s cards' % sCardType)
        sText += '%(num)d %(type)s cards are %(multitype)s cards' \
                ' %(per)s\n' % {'num': iMulti, 'type': sCardType,
                'multitype': sType, 'per': sPer, }
    return sText


def _wrap(sText):
    """Return a gtk.Label which wraps the given text"""
    oLabel = gtk.Label()
    oLabel.set_line_wrap(True)
    oLabel.set_width_chars(80)
    oLabel.set_alignment(0, 0)  # Align top-left
    oLabel.set_markup(sText)
    return oLabel


def _get_back_counts(aPhysCards):
    """Get the counts of the different expansions for the list"""
    dCounts = {}
    for oCard in aPhysCards:
        oKey = oCard.expansion
        if oKey not in ODD_BACKS:
            oKey = 'Other'
        dCounts.setdefault(oKey, 0)
        dCounts[oKey] += 1
    return dCounts


def _split_into_crypt_lib(aPhysCards):
    """Split a list of physical cards into crypt and library"""
    aCrypt = []
    aLib = []
    for oCard in aPhysCards:
        # pylint: disable-msg=E1101
        # pyprotocols confuses pylint
        oAbsCard = IAbstractCard(oCard)
        sType = [x.name for x in oAbsCard.cardtype][0]
        if sType in CRYPT_TYPES:
            aCrypt.append(oCard)
        else:
            aLib.append(oCard)
    return aCrypt, aLib


def _check_same(aPhysCards):
    """Check that all the crypt cards and all the library cards have the
       same backs"""
    aCrypt, aLib = _split_into_crypt_lib(aPhysCards)
    dCrypt = _get_back_counts(aCrypt)
    dLib = _get_back_counts(aLib)
    if len(dCrypt) > 1 or len(dLib) > 1:
        return False  # Differing backs
    return True


def _simple_back_checks(dCards, sType):
    """Simple checks on the card backs. Returns None if we need
       further investigation"""
    if len(dCards) == 1 and not dCards.has_key(None):
        sText = "All %s cards have identical backs\n" % sType.lower() + \
                    UNSLEEVED % sType
        return sText
    elif dCards.has_key(None):
        sText = "%s has cards from unspecified expansions." \
                    " Ignoring the %s\n" % (sType, sType.lower())
        return sText
    return None


def _percentage_backs(dCards, iSize, fPer, sType):
    """Checks that the percentage of cards with a single back are not too
       small"""
    sText = ""
    bOK = True
    for oExp in dCards:
        fThisPer = float(dCards[oExp]) / float(iSize) * 100
        if hasattr(oExp, 'name'):
            sName = oExp.name
        else:
            sName = oExp
        sText += "%d %s backs (%2.1f%% of the %s).\n" % (dCards[oExp], sName,
                fThisPer, sType.lower())
        if fThisPer < fPer:
            sText += "Group of cards with the same back that's smaller" \
                    " than %2.1f%% of the %s.\n" % (fPer, sType.lower())
            bOK = False
    return sText, bOK


def _group_backs(dCards, aCards, iNum):
    """Check for that cards of a single back don't belong to too few different
       card groups"""
    # No more than 2 distinct vampires of in a group of common backs
    dCardsByExp = {}
    bOK = True
    sText = ""
    for oExp in dCards:
        if oExp != 'Other':
            dCardsByExp[oExp] = [oCard for oCard in aCards if
                    oCard.expansion == oExp]
        else:
            dCardsByExp[oExp] = [oCard for oCard in aCards if
                    oCard.expansion not in ODD_BACKS]
    for oExp, aExpCards in dCardsByExp.iteritems():
        # For each expansion, count number of distinct cards
        aNames = set([x.abstractCard.name for x in aExpCards])
        if len(aNames) < iNum:
            sText += "Group of fewer than %d different cards" \
                    " with the same back.\n" % iNum
            sText += "\t" + ", ".join(aNames) + "\n"
            bOK = False
    if not bOK:
        return sText, dCardsByExp
    else:
        return None, dCardsByExp


class DisciplineNumberSelect(gtk.HBox):
    # pylint: disable-msg=R0904
    # gtk.Widget so many public methods
    """Holds a combo box and a discpline list for choosing a list
       of disciplines to use."""

    _sUseList = 'Use list of disciplines'

    def __init__(self, aSortedDisciplines, oDlg):
        super(DisciplineNumberSelect, self).__init__(False, 2)
        self._aSortedDisciplines = aSortedDisciplines
        # Never show more than 5 disciplines here - people can use the
        # discpline list in the combo box if they want more
        self._oComboBox = gtk.combo_box_new_text()
        for iNum in range(1, min(5, len(aSortedDisciplines)) + 1):
            self._oComboBox.append_text(str(iNum))
        self._oComboBox.append_text(self._sUseList)
        self._oComboBox.set_active(1)
        self.pack_start(gtk.Label('Number of Disciplines'))
        self.pack_start(self._oComboBox)
        self.pack_start(gtk.Label(' : '))
        self._oDiscWidget = MultiSelectComboBox(oDlg)
        self._oDiscWidget.fill_list(self._aSortedDisciplines)
        self._oDiscWidget.set_sensitive(False)
        self._oDiscWidget.set_list_size(200, 400)
        self.pack_start(self._oDiscWidget)
        self._oComboBox.connect('changed', self._combo_changed)

    def _combo_changed(self, _oWidget):
        """Toggle the sensitivity of the Discipline select widget as needed"""
        if self._oComboBox.get_active_text() == self._sUseList:
            self._oDiscWidget.set_sensitive(True)
        else:
            self._oDiscWidget.set_sensitive(False)

    def get_disciplines(self):
        """Get the list of disciplines to use."""
        if self._oComboBox.get_active_text() == 'Use list of disciplines':
            aTheseDiscs = self._oDiscWidget.get_selection()
        else:
            iNumDiscs = int(self._oComboBox.get_active_text())
            aTheseDiscs = self._aSortedDisciplines[:iNumDiscs]
        return aTheseDiscs


class AnalyzeCardList(SutekhPlugin):
    """Plugin to analyze card sets.

      Displays various interesting stats, and does a Happy Family
      analysis of the deck
       """
    dTableVersions = {PhysicalCardSet: (6,)}
    aModelsSupported = (PhysicalCardSet,)

    def get_menu_item(self):
        """Register on the 'Analyze' Menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oAnalyze = gtk.MenuItem("Analyze Deck")
        oAnalyze.connect("activate", self.activate)
        return ('Analyze', oAnalyze)

    # pylint: disable-msg=W0201, R0915
    # W0201 - We define a lot of class variables here, because a) this is the
    # plugin entry point, and, b) they need to reflect the current CardSet,
    # so they can't be filled properly in __init__
    # R0915 - This is responsible for filling the whole notebook, so quite
    # a long function, and there's no benefit to splitting it up further
    def activate(self, _oWidget):
        """Create the actual dialog, and populate it"""
        if not self.check_cs_size('Analyze Deck', 500):
            return
        oDlg = SutekhDialog("Analysis of Card List", self.parent,
                gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        oDlg.connect("response", lambda oDlg, resp: oDlg.destroy())
        dConstruct = {
                'Vampire': self._process_vampire,
                'Imbued': self._process_imbued,
                'Combat': self._process_combat,
                'Reaction': self._process_reaction,
                'Action': self._process_action,
                'Action Modifier': self._process_action_modifier,
                'Retainer': self._process_retainer,
                'Equipment': self._process_equipment,
                'Ally': self._process_allies,
                'Political Action': self._process_political_action,
                'Power': self._process_power,
                'Conviction': self._process_conviction,
                'Event': self._process_event,
                'Master': self._process_master,
                'Multirole': self._process_multi,
                'Not Tournament Legal Cards': self._process_non_legal,
                'Mixed Card Backs': self._process_backs,
                }

        self.dTypeNumbers = {}
        dCardLists = {}

        aAllPhysCards = [IPhysicalCard(x) for x in
                self.model.get_card_iterator(None)]
        aAllCards = _get_abstract_cards(aAllPhysCards)

        for sCardType in dConstruct:
            if sCardType not in SPECIAL:
                oFilter = CardTypeFilter(sCardType)
                dCardLists[sCardType] = _get_abstract_cards(
                        self.model.get_card_iterator(oFilter))
                self.dTypeNumbers[sCardType] = len(dCardLists[sCardType])
            elif sCardType == 'Multirole':
                 # Multirole values start empty, and are filled in later
                dCardLists[sCardType] = []
                self.dTypeNumbers[sCardType] = 0
            elif sCardType == 'Not Tournament Legal Cards':
                oFilter = FilterNot(self.model.oLegalFilter)
                dCardLists[sCardType] = _get_abstract_cards(
                        self.model.get_card_iterator(oFilter))
                self.dTypeNumbers[sCardType] = len(dCardLists[sCardType])
            elif sCardType == 'Mixed Card Backs':
                # Assume not mixed, so we skip
                self.dTypeNumbers[sCardType] = 0
                if not _check_same(aAllPhysCards):
                    self.dTypeNumbers[sCardType] = len(aAllPhysCards)
                dCardLists[sCardType] = aAllPhysCards

        oHappyBox = gtk.VBox(False, 2)

        self.iTotNumber = len(aAllCards)
        self.dCryptStats = {}
        self.dLibStats = {}

        self.iCryptSize = sum([self.dTypeNumbers[x] for x in CRYPT_TYPES])
        self.iLibSize = len(aAllCards) - self.iCryptSize
        self.get_crypt_stats(dCardLists['Vampire'], dCardLists['Imbued'])
        self.get_library_stats(aAllCards, dCardLists)
        # Do happy family analysis
        self.happy_families_init(oHappyBox, oDlg)

        # Fill the dialog with the results
        oNotebook = gtk.Notebook()
        # Oh, popup_enable and scrollable - how I adore thee
        oNotebook.set_scrollable(True)
        oNotebook.popup_enable()
        oMainBox = gtk.VBox(False, 2)
        oNotebook.append_page(oMainBox, gtk.Label('Basic Info'))
        oNotebook.append_page(oHappyBox, gtk.Label('Happy Families Analysis'))

        # overly clever? crypt cards first, then alphabetical, then specials
        aOrderToList = sorted(CRYPT_TYPES) + \
                [x for x in sorted(self.dTypeNumbers) if
                        (x not in CRYPT_TYPES and x not in SPECIAL)] + \
                                sorted(SPECIAL)
        for sCardType in aOrderToList:
            if self.dTypeNumbers[sCardType]:
                fProcess = dConstruct[sCardType]
                oNotebook.append_page(_wrap(fProcess(dCardLists[sCardType])),
                        gtk.Label(sCardType))

        # Setup the main notebook
        oMainBox.pack_start(_wrap(self._prepare_main()))
        if self.iLibSize > 0:
            oMainBox.pack_start(self._process_library())
        # pylint: disable-msg=E1101
        # vbox methods not seen
        oDlg.vbox.pack_start(oNotebook)
        oDlg.show_all()
        oNotebook.set_current_page(0)
        oDlg.run()

    # pylint: enable-msg=W0201, R0915

    def get_crypt_stats(self, aVampireCards, aImbuedCards):
        """Extract the relevant statistics about the crypt from the lists
           of cards."""
        def get_info(aVampires, aImbued, sClass):
            """Extract the minimum and maximum for the sets into
               self.dCryptStats, using keys of the form 'vampire min sClass'"""
            sMax = 'max %s' % sClass
            sMin = 'min %s' % sClass
            if len(aImbued):
                iIMax = self.dCryptStats['imbued ' + sMax] = max(aImbued)
                iIMin = self.dCryptStats['imbued ' + sMin] = min(aImbued)
            else:
                iIMax = -500
                iIMin = 500
            if len(aVampires):
                iVMax = self.dCryptStats['vampire ' + sMax] = max(aVampires)
                iVMin = self.dCryptStats['vampire ' + sMin] = min(aVampires)
            else:
                iVMax = -500
                iVMin = 500
            self.dCryptStats[sMax] = max(iVMax, iIMax)
            self.dCryptStats[sMin] = min(iVMin, iIMin)

        # Skip the any group case, as it has no effect here
        get_info([x.group for x in aVampireCards if x.group != -1],
                [x.group for x in aImbuedCards if x.group != -1], 'group')
        get_info([x.capacity for x in aVampireCards],
                [x.life for x in aImbuedCards], 'cost')
        aAllCosts = sorted([x.capacity for x in aVampireCards] + \
                [x.life for x in aImbuedCards])
        self.dCryptStats['total cost'] = sum(aAllCosts)
        self.dCryptStats['min draw'] = sum(aAllCosts[0:4])
        self.dCryptStats['max draw'] = sum(aAllCosts[-1:-5:-1])
        # Extract discipline stats (will be used in display + HF)
        dDiscs = {}
        self.dCryptStats['crypt discipline'] = dDiscs
        for oCard in aVampireCards:
            for oDisc in oCard.discipline:
                dDiscs.setdefault(oDisc.discipline, ['discipline', 0, 0])
                dDiscs[oDisc.discipline][1] += 1
                if oDisc.level == 'superior':
                    dDiscs[oDisc.discipline][2] += 1
        # We treat virtues as inferior discipline for happy family analysis
        for oCard in aImbuedCards:
            for oVirt in oCard.virtue:
                dDiscs.setdefault(oVirt, ['virtue', 0, 0])
                dDiscs[oVirt][1] += 1

    def get_library_stats(self, aAllCards, dCardLists):
        """Extract the relevant library stats from the list of cards"""
        aCryptCards = []
        for sType in CRYPT_TYPES:
            aCryptCards.extend(dCardLists[sType])
        aLibCards = [x for x in aAllCards if x not in aCryptCards]
        # Extract the relevant stats
        self.dLibStats['clan'] = {'No Clan': 0}
        self.dLibStats['discipline'] = {'No Discipline': 0}
        for oCard in aLibCards:
            if len(oCard.cardtype) > 1:
                self.dTypeNumbers['Multirole'] += 1
                dCardLists['Multirole'].append(oCard)
            if len(oCard.clan) > 0:
                aClans = [x.name for x in oCard.clan]
            elif len(oCard.creed) > 0:
                aClans = [x.name for x in oCard.creed]
            else:
                aClans = ['No Clan']
            for sClan in aClans:
                self.dLibStats['clan'].setdefault(sClan, 0)
                self.dLibStats['clan'][sClan] += 1
            if len(oCard.discipline) > 0:
                aDisciplines = [oP.discipline.fullname for oP in
                        oCard.discipline]
            elif len(oCard.virtue) > 0:
                aDisciplines = [oV.fullname for oV in oCard.virtue]
            else:
                aDisciplines = ['No Discipline']
            for sDisc in aDisciplines:
                self.dLibStats['discipline'].setdefault(sDisc, 0)
                self.dLibStats['discipline'][sDisc] += 1

    def _prepare_main(self):
        """Setup the main notebook display"""
        oCS = self.get_card_set()

        sMainText = "Analysis Results for :\n\t\t<b>%(name)s</b>\n" \
                "\t\tby <i>%(author)s</i>\n\t\tDescription: <i>%(desc)s</i>" \
                "\n\n" % {
                        'name': self.escape(self.view.sSetName),
                        'author': self.escape(oCS.author),
                        'desc': self.escape(oCS.comment),
                        }

        # Set main notebook text
        for sCardType in CRYPT_TYPES:
            if self.dTypeNumbers[sCardType] > 0:
                sMainText += 'Number of %s = %d\n' % (sCardType,
                        self.dTypeNumbers[sCardType])
        if (self.dTypeNumbers['Vampire'] > 0 and
                self.dTypeNumbers['Imbued'] > 0) or self.iCryptSize == 0:
            sMainText += "Total Crypt size = %d\n" % self.iCryptSize
        if self.iCryptSize > 0:
            sMainText += "Minimum Group in Crypt = %d\n" % \
                    self.dCryptStats['min group']
            sMainText += "Maximum Group in Crypt = %d\n" % \
                    self.dCryptStats['max group']

        if self.iCryptSize < 12:
            sMainText += '<span foreground = "red">Less than 12 Crypt Cards' \
                    '</span>\n'

        if self.dCryptStats['max group'] - self.dCryptStats['min group'] > 1:
            sMainText += '<span foreground = "red">Group Range Exceeded' \
                    '</span>\n'

        if self.dTypeNumbers['Not Tournament Legal Cards'] > 0:
            sMainText += '<span foreground = "red">Card Set uses cards that ' \
                    'are not legal for tournament play</span>\n'

        if self.iCryptSize > 0:
            sMainText += '\nMaximum cost in crypt = %d\n' % \
                    self.dCryptStats['max cost']
            sMainText += 'Minimum cost in crypt = %d\n' % \
                    self.dCryptStats['min cost']
        fAvg, fAvgDraw = _crypt_avg(self.dCryptStats['total cost'],
                self.iCryptSize)

        sMainText += 'Average cost = %3.2f (%3.2f average crypt draw cost)\n' \
                % (fAvg, fAvgDraw)
        sMainText += 'Minimum draw cost = %d\n' % self.dCryptStats['min draw']
        sMainText += 'Maximum Draw cost = %d\n' % self.dCryptStats['max draw']

        sMainText += "Total Library Size = %d\n" % self.iLibSize

        if self.iLibSize < 40:
            sMainText += '<span foreground = "red">Less than 40 Library' \
                    ' Cards</span>\n'
        elif self.iLibSize < 60:
            sMainText += '<span foreground = "orange">Less than 60 Library' \
                    ' Cards - this deck is not legal for standard' \
                    ' constructed tournaments</span>'
        elif self.iLibSize > 90:
            sMainText += '<span foreground = "orange">More than 90 Library' \
                    ' Cards - this deck is not legal for standard' \
                    ' constructed tournaments</span>\n'

        return sMainText

    def _process_library(self):
        """Create a notebook for the basic library card overview"""
        oLibNotebook = gtk.Notebook()
        # Stats by card type
        sTypeText = ''
        # Show card types, sorted by number (then alphabetical by type)
        for sType, iCount in sorted(self.dTypeNumbers.items(),
                key=lambda x: (x[1], x[0]), reverse=True):
            if sType not in CRYPT_TYPES and sType not in SPECIAL and \
                    iCount > 0:
                sTypeText += _format_card_line(sType, 'cards', iCount,
                        self.iLibSize)
        if self.dTypeNumbers['Multirole'] > 0:
            sTypeText += '\n' + _format_card_line('Multirole', 'cards',
                    self.dTypeNumbers['Multirole'], self.iLibSize)
        oLibNotebook.append_page(_wrap(sTypeText),
                gtk.Label('Card Types'))
        # Stats by discipline
        sDiscText = _format_card_line('Master', 'cards',
                self.dTypeNumbers['Master'], self.iLibSize)
        sDiscText += _format_card_line('non-master cards with No '
                'discipline requirement', '',
                self.dLibStats['discipline']['No Discipline'],
                self.iLibSize) + '\n'
        # sort by number, then name
        for sDisc, iNum in sorted(self.dLibStats['discipline'].items(),
                key=lambda x: (x[1], x[0]), reverse=True):
            if sDisc != 'No Discipline' and iNum > 0:
                sDiscDesc = 'non-master cards with %s' % sDisc
                sDiscText += _format_card_line(sDiscDesc, '', iNum,
                        self.iLibSize)
        oLibNotebook.append_page(_wrap(sDiscText),
                gtk.Label('Disciplines'))
        # Stats by clan requirement
        sClanText = _format_card_line('cards with No clan requirement', '',
                self.dLibStats['clan']['No Clan'], self.iLibSize) + '\n'
        for sClan, iNum in sorted(self.dLibStats['clan'].items()):
            if sClan != 'No Clan' and iNum > 0:
                sClanDesc = 'cards requiring %s' % sClan
                sClanText += _format_card_line(sClanDesc, '', iNum,
                        self.iLibSize)
        oLibNotebook.append_page(_wrap(sClanText),
                gtk.Label('Clan Requirements'))
        return oLibNotebook

    def _process_vampire(self, aCards):
        """Process the list of vampires"""
        dDeckDetails = {
                'vampires': {},
                'titles': {},
                'clans': {},
                'sects': {},
                'votes': 0,
                }
        iNum = self.dTypeNumbers['Vampire']
        for oAbsCard in aCards:
            dDeckDetails['vampires'].setdefault(oAbsCard.name, 0)
            dDeckDetails['vampires'][oAbsCard.name] += 1
            for oClan in oAbsCard.clan:
                dDeckDetails['clans'].setdefault(oClan, 0)
                dDeckDetails['clans'][oClan] += 1
            for oSect in oAbsCard.sect:
                dDeckDetails['sects'].setdefault(oSect, 0)
                dDeckDetails['sects'][oSect] += 1
            for oTitle in oAbsCard.title:
                dDeckDetails['titles'].setdefault(oTitle, 0)
                dDeckDetails['titles'][oTitle] += 1
                dDeckDetails['votes'] += Titles.vote_value(oTitle.name)
        # Build up Text
        sVampText = "\t\t<b>Vampires :</b>\n\n"
        sVampText += '<span foreground = "blue">Basic Crypt stats</span>\n'
        sVampText += "Number of Vampires = %d %s\n" % (iNum,
                _percentage(iNum, self.iCryptSize, "Crypt"))
        sVampText += "Number of Unique Vampires = %d\n" % len(
                dDeckDetails['vampires'])
        sVampText += "Minimum Group is : %d\n" % \
                self.dCryptStats['vampire min group']
        sVampText += "Maximum Group is : %d\n" % \
                self.dCryptStats['vampire max group']
        sVampText += '\n<span foreground = "blue">Crypt cost</span>\n'
        sVampText += "Cheapest is : %d\n" % \
                self.dCryptStats['vampire min cost']
        sVampText += "Most Expensive is : %d\n" % \
                self.dCryptStats['vampire max cost']
        sVampText += "Average Capacity is : %2.3f\n\n" % (
                sum([x.capacity for x in aCards]) / float(iNum))
        sVampText += '<span foreground = "blue">Clans</span>\n'
        for oClan, iCount in dDeckDetails['clans'].iteritems():
            sVampText += "%d Vampires of clan %s %s\n" % (iCount,
                    oClan.name, _percentage(iCount, self.iCryptSize, "Crypt"))
        sVampText += '<span foreground = "blue">Sects</span>\n'
        for oSect, iCount in dDeckDetails['sects'].iteritems():
            sVampText += "%d %s vampires %s\n" % (iCount,
                    oSect.name, _percentage(iCount, self.iCryptSize, "Crypt"))
        sVampText += '\n<span foreground = "blue">Titles</span>\n'
        iTotalTitles = 0
        for oTitle, iCount in dDeckDetails['titles'].iteritems():
            sVampText += "%d vampires with the title %s (%d votes)\n" % (
                    iCount, oTitle.name, Titles.vote_value(oTitle.name))
            iTotalTitles += iCount
        sVampText += "%d vampires with titles (%s)\n" % (
                iTotalTitles, _percentage(iTotalTitles, self.iCryptSize,
                    "Crypt"))
        sVampText += "%d votes from titles in the crypt. Average votes per" \
                " vampire is %2.3f\n" % (dDeckDetails['votes'],
                        dDeckDetails['votes'] / float(iNum))
        sVampText += '\n<span foreground = "blue">Disciplines</span>\n'
        for oDisc, aInfo in sorted(
                self.dCryptStats['crypt discipline'].iteritems(),
                cmp=_disc_sort_cmp):
            if aInfo[0] == 'discipline':
                sVampText += "%(infcount)d Vampires with %(disc)s %(iper)s," \
                        " %(supcount)d at Superior %(sper)s\n" % {
                                'disc': oDisc.fullname,
                                'infcount': aInfo[1],
                                'iper': _percentage(aInfo[1],
                                    self.iCryptSize, "Crypt"),
                                'supcount': aInfo[2],
                                'sper': _percentage(aInfo[2],
                                    self.iCryptSize, "Crypt"),
                                }
        return sVampText

    def _process_imbued(self, aCards):
        """Fill the Imbued tab"""
        dDeckImbued = {}
        dDeckCreed = {}
        iNum = self.dTypeNumbers['Imbued']
        for oAbsCard in aCards:
            dDeckImbued.setdefault(oAbsCard.name, 0)
            dDeckImbued[oAbsCard.name] += 1
            for oCreed in oAbsCard.creed:
                dDeckCreed.setdefault(oCreed, 0)
                dDeckCreed[oCreed] += 1
        # Build up Text
        sImbuedText = "\t\t<b>Imbued</b>\n\n"
        sImbuedText += '<span foreground = "blue">Basic Crypt stats</span>\n'
        sImbuedText += "Number of Imbued = %d %s\n" % (iNum, _percentage(iNum,
            self.iCryptSize, "Crypt"))
        sImbuedText += "Number of Unique Imbued = %d\n" % len(dDeckImbued)
        sImbuedText += 'Minimum Group is : %d\n' % \
                self.dCryptStats['imbued min group']
        sImbuedText += 'Maximum Group is : %d\n' % \
                self.dCryptStats['imbued max group']
        sImbuedText += '\n<span foreground = "blue">Crypt cost</span>\n'
        sImbuedText += "Cheapest is : %d\n" % \
                self.dCryptStats['imbued min cost']
        sImbuedText += "Most Expensive is : %d\n" % \
                self.dCryptStats['imbued max cost']
        sImbuedText += "Average Life is : %2.3f\n\n" % (
                sum([x.life for x in aCards]) / float(iNum))
        for oCreed, iCount in dDeckCreed.iteritems():
            sImbuedText += "%d Imbued of creed %s %s\n" % (iCount,
                    oCreed.name, _percentage(iCount, self.iCryptSize, "Crypt"))
        for oVirtue, aInfo in sorted(
                self.dCryptStats['crypt discipline'].iteritems(),
                cmp=_disc_sort_cmp):
            if aInfo[0] == 'virtue':
                sImbuedText += "%d Imbued with %s %s\n" % (aInfo[1],
                        oVirtue.fullname, _percentage(aInfo[1],
                            self.iCryptSize, "Crypt"))
        return sImbuedText

    def _process_master(self, aCards):
        """Display the stats for Master Cards"""
        iNum = self.dTypeNumbers['Master']
        _aBlood, aPool, _aConviction = _get_card_costs(aCards)
        iClanRequirement, dClan, dMulti = _get_card_clan_multi(aCards)
        # Build up Text
        sText = "\t\t<b>Master Cards :</b>\n\n"
        sText += "Number of Masters = %d %s\n" % (iNum,
                _percentage(iNum, self.iLibSize, "Library"))
        # Extract the number of out-of-turn and trifles
        for sType in ('trifle', 'out-of-turn'):
            try:
                oKeyword = IKeyword(sType)
                iCount = 0
                for oCard in aCards:
                    if oKeyword in oCard.keywords:
                        iCount += 1
                if iCount:
                    sText += '   Number of %s masters = %d (%s)\n' % (sType,
                            iCount, _percentage(iCount, iNum, 'Masters'))
            except SQLObjectNotFound:
                sText += '<span foreground="red">Keyword <b>%s</b> not '\
                        'in the database -- please re-import the WW ' \
                        'cardlist</span>\n' % sType
        if aPool[1] > 0:
            sText += '\n<span foreground = "blue">Cost</span>\n'
            sText += _format_cost_numbers('Master', 'pool', aPool, iNum)
        if iClanRequirement > 0:
            sText += '\n<span foreground = "blue">Clan/Creed</span>\n'
            sText += _format_clan('Master', iClanRequirement, dClan, iNum)
        if len(dMulti) > 0:
            sText += '\n' + _format_multi('Master', dMulti, iNum)
        return sText

    def _default_text(self, aCards, sType):
        """Standard boilerplate for most card types"""
        iNum = self.dTypeNumbers[sType]
        aBlood, aPool, aConviction = _get_card_costs(aCards)
        iClanRequirement = 0
        dDisciplines, dVirtues, iNoneCount = _get_card_disciplines(aCards)
        iClanRequirement, dClan, dMulti = _get_card_clan_multi(aCards)
        # Build up Text
        sPerCards = _percentage(iNum, self.iLibSize, 'Library')
        sText = "\t\t<b>%(type)s Cards :</b>\n\n" \
                "Number of %(type)s cards = %(num)d %(per)s\n" % {
                        'type': sType,
                        'num': iNum,
                        'per': sPerCards, }
        if aBlood[1] > 0 or aPool[1] > 0 or aConviction[1] > 0:
            sText += '\n<span foreground = "blue">Costs</span>\n'
        if aBlood[1] > 0:
            sText += _format_cost_numbers(sType, 'blood', aBlood, iNum)
        if aConviction[1] > 0:
            sText += _format_cost_numbers(sType, 'conviction', aConviction,
                    iNum)
        if aPool[1] > 0:
            sText += _format_cost_numbers(sType, 'pool', aPool, iNum)
        if iClanRequirement > 0:
            sText += '\n<span foreground = "blue">Clan/Creed</span>\n'
            sText += _format_clan(sType, iClanRequirement, dClan, iNum)
        sText += '\n<span foreground = "blue">Discipline/Virtues</span>\n'
        sText += 'Number of cards with no discipline/virtue requirement = ' \
                '%d\n' % iNoneCount
        if len(dDisciplines) > 0:
            sText += _format_disciplines('discipline', dDisciplines, iNum)
        if len(dVirtues) > 0:
            sText += _format_disciplines('virtue', dVirtues, iNum)
        if len(dMulti) > 0:
            sText += '\n' + _format_multi(sType, dMulti, iNum)
        return sText

    def _process_combat(self, aCards):
        """Fill the combat tab"""
        sText = self._default_text(aCards, 'Combat')
        return sText

    def _process_action_modifier(self, aCards):
        """Fill the Action Modifier tab"""
        sText = self._default_text(aCards, 'Action Modifier')
        return sText

    def _process_reaction(self, aCards):
        """Fill the reaction tab"""
        sText = self._default_text(aCards, 'Reaction')
        return sText

    def _process_event(self, aCards):
        """Fill the events tab"""
        iNumEvents = len(aCards)
        sEventText = "\t\t<b>Event Cards :</b>\n\n"
        sEventText += "Number of Event cards = %d %s\n\n" % (iNumEvents,
                _percentage(iNumEvents, self.iLibSize, "Library"))
        sEventText += '<span foreground = "blue">Event classes</span>\n'
        dEventTypes = {}
        for oCard in aCards:
            sType = oCard.text.split('.', 1)[0]  # first word is type
            dEventTypes.setdefault(sType, 0)
            dEventTypes[sType] += 1
        for sType, iCount in dEventTypes.iteritems():
            sEventText += '%d of type %s : %s (%s) \n' % (iCount, sType,
                    _percentage(iCount, iNumEvents, 'Events'),
                    _percentage(iCount, self.iLibSize, 'Library'))
        return sEventText

    def _process_action(self, aCards):
        """Fill the actions tab"""
        sText = self._default_text(aCards, 'Action')
        return sText

    def _process_political_action(self, aCards):
        """Fill the Political Actions tab"""
        sText = self._default_text(aCards, 'Political Action')
        return sText

    def _process_allies(self, aCards):
        """Fill the allies tab"""
        sText = self._default_text(aCards, 'Ally')
        return sText

    def _process_retainer(self, aCards):
        """Fill the retainer tab"""
        sText = self._default_text(aCards, 'Retainer')
        return sText

    def _process_equipment(self, aCards):
        """Fill the equipment tab"""
        sText = self._default_text(aCards, 'Equipment')
        return sText

    def _process_conviction(self, aCards):
        """Fill the conviction tab"""
        sText = self._default_text(aCards, 'Conviction')
        return sText

    def _process_power(self, aCards):
        """Fill the power tab"""
        sText = self._default_text(aCards, 'Power')
        return sText

    def _process_multi(self, aCards):
        """Fill the multirole card tab"""
        dMulti = {}
        sPerCards = _percentage(self.dTypeNumbers['Multirole'], self.iLibSize,
                'Library')
        sText = "\t\t<b>Multirole Cards :</b>\n\n" \
                "Number of Multirole cards = %(num)d %(per)s\n" % {
                        'num': self.dTypeNumbers['Multirole'],
                        'per': sPerCards,
                        }
        for oAbsCard in aCards:
            aTypes = [x.name for x in oAbsCard.cardtype]
            if len(aTypes) > 1:
                sKey = "/".join(sorted(aTypes))
                dMulti.setdefault(sKey, 0)
                dMulti[sKey] += 1
        for sMultiType, iNum in sorted(dMulti.items(), key=lambda x: x[1],
                reverse=True):
            sPer = _percentage(iNum, self.iLibSize, 'Library')
            sText += 'Number of %(multitype)s cards = %(num)d %(per)s\n' % {
                    'multitype': sMultiType,
                    'num': iNum,
                    'per': sPer,
                    }

        return sText

    def _process_non_legal(self, aCards):
        """Fill the non_legal card tab"""
        iTotal = self.iCryptSize + self.iLibSize
        dNonLegal = {}
        sPerCards = _percentage(
                self.dTypeNumbers['Not Tournament Legal Cards'],
                iTotal, 'Deck')
        sText = "\t\t<b>Not Tournament Legal Cards :</b>\n\n" \
                "Number of cards not legal for tournament play =" \
                " %(num)d %(per)s\n" % {
                        'num': self.dTypeNumbers[
                            'Not Tournament Legal Cards'],
                        'per': sPerCards,
                        }
        for oAbsCard in aCards:
            dNonLegal.setdefault(oAbsCard.name, 0)
            dNonLegal[oAbsCard.name] += 1
        for sName, iNum in sorted(dNonLegal.items(), key=lambda x: x[1],
                reverse=True):
            sText += '%(num)d X %(name)s\n' % {
                    'num': iNum,
                    'name': sName,
                    }

        return sText

    def _check_crypt_backs(self, dCrypt, aCrypt):
        """Check the backs on the crypt cards"""
        sText = "\t<b>Crypt</b>\n\n"
        sAddText = _simple_back_checks(dCrypt, 'Crypt')
        bOK = True
        if sAddText:
            sText += sAddText
            return sText
        sAddText, bOK = _percentage_backs(dCrypt, self.iCryptSize, 20.0,
                'Crypt')
        if sAddText:
            sText += sAddText
        sAddText, _dCryptByExp = _group_backs(dCrypt, aCrypt, 3)
        if sAddText:
            sText += sAddText
            bOK = False
        if bOK:
            sText += "Mixed backs, but seems sufficiently mixed.\n" \
                    + UNSLEEVED % "Crypt"
        else:
            sText += SLEEVED % "Crypt"

        return sText

    def _check_lib_backs(self, dLib, aLib):
        """Check the card backs for the library"""
        sText = "\n\t<b>Library</b>\n\n"
        sAddText = _simple_back_checks(dLib, 'Library')
        bOK = True
        if sAddText:
            sText += sAddText
            return sText
        sAddText, bOK = _percentage_backs(dLib, self.iLibSize, 20.0, 'Library')
        if sAddText:
            sText += sAddText
        sAddText, dLibByExp = _group_backs(dLib, aLib, 5)
        if sAddText:
            sText += sAddText
            bOK = False
        for aExpCards in dLibByExp.itervalues():
            # if a single back contains less than 3 types of library card
            aAllTypes = set()
            for oCard in aExpCards:
                aTypes = [y.name for y in oCard.abstractCard.cardtype]
                aAllTypes.add("/".join(aTypes))
            if len(aAllTypes) < 3:
                aNames = set([x.abstractCard.name for x in aExpCards])
                sText += "Group of cards with the same back" \
                        " containing less than 3 different card types.\n"
                sText += "\t" + ", ".join(aNames) + "\n"
                bOK = False
        if bOK:
            sText += "Mixed backs, but seems sufficiently mixed.\n" + \
                    UNSLEEVED % "Library"
        else:
            sText += SLEEVED % "Library"

        return sText

    def _process_backs(self, aPhysCards):
        """Run some heuristic tests to see if cards are 'of sufficiently
           mixed card type'"""
        aCrypt, aLib = _split_into_crypt_lib(aPhysCards)
        dCrypt = _get_back_counts(aCrypt)
        dLib = _get_back_counts(aLib)
        sText = "\t\t<b>Mixed Card Backs :</b>\n\n"
        if dCrypt.has_key(None) and dLib.has_key(None):
            sText += "Both library and crypt have cards with unspecified" \
                    " expansions.\nIgnoring the mixed backs.\n"
            return sText
        # We know there are mixed cards, when we get here.
        sText += "Mixed card backs in the deck. The V:EKN rules require " \
                "that a mixed deck be of 'sufficiently mixed card types' " \
                "if it's to be played without sleeves. This tests some " \
                "obvious cases, but check with the event judge if playing " \
                "without sleeves\n\n"
        sText += self._check_crypt_backs(dCrypt, aCrypt)
        sText += self._check_lib_backs(dLib, aLib)

        return sText

    def happy_families_init(self, oHFVBox, oDlg):
        """Setup data and tab for HF analysis"""
        oMainLabel = gtk.Label()
        oHFVBox.pack_start(oMainLabel)
        # Build up Text
        sHappyFamilyText = "\t\t<b>Happy Families Analysis :</b>\n"
        sHappyFamilyText += "\t<small><i>Happy Families Analysis from the " \
                "Gangrel antitribu newsletter, Jan 2000, by Legbiter" \
                "</i></small>\n"
        if self.dTypeNumbers['Imbued'] > 0:
            sHappyFamilyText += '\n<span foreground = "red">This is not' \
                    ' optimised for Imbued, and treats them as small ' \
                    'vampires</span>\n'
        if self.dTypeNumbers['Event'] > 0:
            sHappyFamilyText += '\n<span foreground = "red">Events are' \
                    ' not considered by Happy Families. We treat them' \
                    ' as additional masters.</span>'
        if self.iCryptSize == 0:
            sHappyFamilyText += '\n<span foreground = "red">Need to have' \
                    ' a crypt to do the analysis</span>\n'
            oMainLabel.set_markup(sHappyFamilyText)
            oHFVBox.show_all()
            return
        if len(self.dCryptStats['crypt discipline']) < 1:
            # Crypt only has Smudge, for example
            sHappyFamilyText += '\n<span foreground = "red">Need disciplines' \
                    ' in the crypt to do analysis</span>\n'
            oMainLabel.set_markup(sHappyFamilyText)
            oHFVBox.show_all()
            return
        # OK, for analysis, so set eveything up
        # Masters analysis
        iHFMasters = int(round(0.2 * self.iLibSize))
        iLibMasters = self.dTypeNumbers['Master'] + self.dTypeNumbers['Event']
        iNonMasters = (self.iLibSize - self.dTypeNumbers['Master']
                - self.dTypeNumbers['Event'])
        sHappyFamilyText += "\n\t<b>Master Cards</b>\n"
        sHappyFamilyText += "%d Masters %s\n" % (self.dTypeNumbers['Master'],
                _percentage(self.dTypeNumbers['Master'],
                        self.iLibSize, "Library"))
        if self.dTypeNumbers['Event'] > 0:
            sHappyFamilyText += "%d Events %s\n" % (self.dTypeNumbers['Event'],
                    _percentage(self.dTypeNumbers['Event'], self.iLibSize,
                        "Library"))
            sHappyFamilyText += "%d Total %s\n" % (iLibMasters,
                    _percentage(iLibMasters, self.iLibSize, 'Library'))
        sHappyFamilyText += "Happy Families recommends 20%, which would" \
                " be " + str(iHFMasters) + '  : '
        sHappyFamilyText += "<span foreground = \"blue\">Difference = " + \
                str(abs(iHFMasters - iLibMasters)) + \
                "</span>\n"
        # Discipline analysis
        aSortedDiscs = [x[0].fullname for x in sorted(
            self.dCryptStats['crypt discipline'].items(), cmp=_disc_sort_cmp)]
        oMainLabel.set_markup(sHappyFamilyText)
        oDiscSelect = DisciplineNumberSelect(aSortedDiscs, oDlg)
        oHFVBox.pack_start(oDiscSelect, False, False)
        oResLabel = gtk.Label()
        oButton = gtk.Button('Recalculate Happy Family Analysis')
        oButton.connect('clicked', self._redo_happy_family, oDiscSelect,
                oResLabel)
        oHFVBox.pack_start(oButton, False, False)
        oResLabel.set_markup(self._happy_lib_analysis(aSortedDiscs[:2],
            iNonMasters))
        oHFVBox.pack_start(oResLabel)
        oHFVBox.show_all()

    def _redo_happy_family(self, _oButton, oDiscSelect, oResLabel):
        """Redo the HF analysis based on button press"""
        aTheseDiscs = oDiscSelect.get_disciplines()
        if not aTheseDiscs:
            return  # Just ignore the zero selection case
        iNonMasters = self.iLibSize - self.dTypeNumbers['Master']
        oResLabel.hide()
        oResLabel.set_markup(self._happy_lib_analysis(aTheseDiscs,
            iNonMasters))
        oResLabel.show()
        oResLabel.get_parent().show_all()

    def _happy_lib_analysis(self, aDiscsToUse, iNonMasters):
        """Heavy lifting of the HF analysis"""

        iNumberToShow = len(aDiscsToUse)

        sHappyFamilyText = "<b>" + str(iNumberToShow) + \
                " Discipline Case</b>\n"
        sHappyFamilyText += "Disciplines : <i>%s</i>\n" % ",".join(aDiscsToUse)
        fDemon = float(self.iCryptSize)
        dCryptDiscs = {}
        for sDisc in aDiscsToUse:
            oDisc = _lookup_discipline(sDisc,
                    self.dCryptStats['crypt discipline'])
            dCryptDiscs[sDisc] = self.dCryptStats['crypt discipline'][oDisc][1]
            fDemon += dCryptDiscs[sDisc]
        iHFNoDiscipline = int((iNonMasters * self.iCryptSize / fDemon))
        iDiff = iNonMasters - iHFNoDiscipline
        dDiscNumbers = {}
        for sDisc in aDiscsToUse:
            iHFNumber = int(iNonMasters * dCryptDiscs[sDisc] / fDemon)
            dDiscNumbers[sDisc] = iHFNumber
            iDiff -= iHFNumber
        if iDiff > 0:
            iHFNoDiscipline += iDiff  # Shove rounding errors here
        sHappyFamilyText += "Number of Minion cards requiring " \
                " No discipline : %s\n" \
                % (self.dLibStats['discipline']['No Discipline']
                        - self.dTypeNumbers['Master']
                        - self.dTypeNumbers['Event'])
        sHappyFamilyText += "Happy Families recommends %d (%.1f %%): " % (
                iHFNoDiscipline, (80 * self.iCryptSize / fDemon))
        sHappyFamilyText += '<span foreground = "blue">Difference = ' \
                '%s</span>\n\n' % abs(iHFNoDiscipline -
                        (self.dLibStats['discipline']['No Discipline']
                            - self.dTypeNumbers['Master']
                            - self.dTypeNumbers['Event']))
        for sDisc in aDiscsToUse:
            iHFNum = dDiscNumbers[sDisc]
            if self.dLibStats['discipline'].has_key(sDisc):
                iLibNum = self.dLibStats['discipline'][sDisc]
            else:
                iLibNum = 0
            sHappyFamilyText += "Number of Minion cards requiring %(disc)s :" \
                    " %(lib)d (%(crypt)d crypt members)\n" % {
                            'disc': sDisc,
                            'lib': iLibNum,
                            'crypt': dCryptDiscs[sDisc],
                            }
            sHappyFamilyText += "Happy Families recommends %d (%.1f %%): " % (
                    iHFNum, 80 * dCryptDiscs[sDisc] / fDemon)
            sHappyFamilyText += '<span foreground = "blue">Difference = ' \
                    '%d </span>\n' % abs(iHFNum - iLibNum)
        return sHappyFamilyText

plugin = AnalyzeCardList
